"""Generate the C# data structures from the intermediate representation."""
import io
import textwrap
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    cast,
    Union,
    Mapping,
    Final,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen import specific_implementations
from aas_core_codegen.common import Error, Identifier, assert_never, Stripped, Rstripped
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
    unrolling as csharp_unrolling,
    description as csharp_description,
)
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II
from aas_core_codegen.intermediate import (
    construction as intermediate_construction,
)


# region Checks


def _human_readable_identifier(
    something: Union[
        intermediate.Enumeration, intermediate.ConcreteClass, intermediate.Interface
    ]
) -> str:
    """
    Represent ``something`` in a human-readable text.

    The reader should be able to trace ``something`` back to the meta-model.\
    """
    result = None  # type: Optional[str]

    if isinstance(something, intermediate.Enumeration):
        result = f"meta-model enumeration {something.name!r}"
    elif isinstance(something, intermediate.ConcreteClass):
        result = f"meta-model class {something.name!r}"
    elif isinstance(something, intermediate.Interface):
        result = f"interface based on the meta-model class {something.name!r}"
    else:
        assert_never(something)

    assert result is not None
    return result


def _verify_structure_name_collisions(
    symbol_table: intermediate.SymbolTable,
) -> List[Error]:
    """Verify that the C# names of the structures do not collide."""
    observed_structure_names: Dict[
        Identifier,
        Union[
            intermediate.Enumeration, intermediate.ConcreteClass, intermediate.Interface
        ],
    ] = dict()

    errors = []  # type: List[Error]

    # region Inter-structure collisions

    for something in csharp_common.over_enumerations_classes_and_interfaces(
        symbol_table
    ):
        name = csharp_naming.name_of(something)

        other = observed_structure_names.get(name, None)

        if other is not None:
            errors.append(
                Error(
                    something.parsed.node,
                    f"The C# name {name!r} "
                    f"of the {_human_readable_identifier(something)} "
                    f"collides with the C# name "
                    f"of the {_human_readable_identifier(other)}",
                )
            )
        else:
            observed_structure_names[name] = something

    # endregion

    # region Intra-structure collisions

    for symbol in symbol_table.symbols:
        collision_error = _verify_intra_structure_collisions(intermediate_symbol=symbol)

        if collision_error is not None:
            errors.append(collision_error)

    # endregion

    return errors


def _verify_intra_structure_collisions(
    intermediate_symbol: intermediate.Symbol,
) -> Optional[Error]:
    """Verify that no member names collide in the C# structure of the given symbol."""
    errors = []  # type: List[Error]

    if isinstance(intermediate_symbol, intermediate.Enumeration):
        pass

    elif isinstance(intermediate_symbol, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(intermediate_symbol, intermediate.Class):
        observed_member_names = {}  # type: Dict[Identifier, str]

        for prop in intermediate_symbol.properties:
            prop_name = csharp_naming.property_name(prop.name)
            if prop_name in observed_member_names:
                # BEFORE-RELEASE (mristin, 2021-12-13): test
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"C# property {prop_name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[prop_name]}",
                    )
                )
            else:
                observed_member_names[prop_name] = (
                    f"C# property {prop_name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

        for method in intermediate_symbol.methods:
            method_name = csharp_naming.method_name(method.name)

            if method_name in observed_member_names:
                # BEFORE-RELEASE (mristin, 2021-12-13): test
                errors.append(
                    Error(
                        method.parsed.node,
                        f"C# method {method_name!r} corresponding "
                        f"to the meta-model method {method.name!r} collides with "
                        f"the {observed_member_names[method_name]}",
                    )
                )
            else:
                observed_member_names[method_name] = (
                    f"C# method {method_name!r} corresponding to "
                    f"the meta-model method {method.name!r}"
                )

    else:
        assert_never(intermediate_symbol)

    if len(errors) > 0:
        errors.append(
            Error(
                intermediate_symbol.parsed.node,
                f"Naming collision(s) in C# code "
                f"for the symbol {intermediate_symbol.name!r}",
                underlying=errors,
            )
        )

    return None


class VerifiedIntermediateSymbolTable(intermediate.SymbolTable):
    """Represent a verified symbol table which can be used for code generation."""

    # noinspection PyInitNewSignature
    def __new__(
        cls, symbol_table: intermediate.SymbolTable
    ) -> "VerifiedIntermediateSymbolTable":
        raise AssertionError("Only for type annotation")


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def verify(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[VerifiedIntermediateSymbolTable], Optional[List[Error]]]:
    """Verify that C# code can be generated from the ``symbol_table``."""
    errors = []  # type: List[Error]

    structure_name_collisions = _verify_structure_name_collisions(
        symbol_table=symbol_table
    )

    errors.extend(structure_name_collisions)

    if len(errors) > 0:
        return None, errors

    return cast(VerifiedIntermediateSymbolTable, symbol_table), None


# endregion

# region Generation


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_enum(
    enum: intermediate.Enumeration,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the C# code for the enum."""
    writer = io.StringIO()

    if enum.description is not None:
        comment, comment_errors = csharp_description.generate_symbol_comment(
            enum.description
        )
        if comment_errors:
            return None, Error(
                enum.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    name = csharp_naming.enum_name(enum.name)
    if len(enum.literals) == 0:
        writer.write(f"public enum {name}\n{{\n}}")
        return Stripped(writer.getvalue()), None

    writer.write(f"public enum {name}\n{{\n")
    for i, literal in enumerate(enum.literals):
        if i > 0:
            writer.write(",\n\n")

        if literal.description:
            (
                literal_comment,
                literal_comment_errors,
            ) = csharp_description.generate_enumeration_literal_comment(
                literal.description
            )

            if literal_comment_errors:
                return None, Error(
                    literal.description.parsed.node,
                    f"Failed to generate the comment "
                    f"for the enumeration literal {literal.name!r}",
                    literal_comment_errors,
                )

            assert literal_comment is not None

            writer.write(textwrap.indent(literal_comment, I))
            writer.write("\n")

        writer.write(
            textwrap.indent(
                f"[EnumMember(Value = {csharp_common.string_literal(literal.value)})]\n"
                f"{csharp_naming.enum_literal_name(literal.name)}",
                I,
            )
        )

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_interface(
    interface: intermediate.Interface,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate C# code for the given interface."""
    writer = io.StringIO()

    if interface.description is not None:
        comment, comment_errors = csharp_description.generate_symbol_comment(
            interface.description
        )

        if comment_errors is not None:
            return None, Error(
                interface.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    name = csharp_naming.interface_name(interface.name)

    inheritances = [inheritance.name for inheritance in interface.inheritances] + [
        Identifier("Class")
    ]

    inheritance_names = list(map(csharp_naming.interface_name, inheritances))

    assert len(inheritances) > 0
    if len(inheritances) == 1:
        writer.write(f"public interface {name} : {inheritance_names[0]}\n{{\n")
    else:
        writer.write(f"public interface {name} :\n")
        for i, inheritance_name in enumerate(inheritance_names):
            if i > 0:
                writer.write(",\n")

            writer.write(textwrap.indent(inheritance_name, II))

        writer.write("\n{\n")

    # Code blocks separated by double newlines and indented once
    blocks = []  # type: List[Stripped]

    # region Getters and setters

    for prop in interface.properties:
        if prop.specified_for is interface.base:
            prop_type = csharp_common.generate_type(
                type_annotation=prop.type_annotation
            )
            prop_name = csharp_naming.property_name(prop.name)

            if prop.description is not None:
                (
                    prop_comment,
                    prop_comment_errors,
                ) = csharp_description.generate_property_comment(prop.description)

                if prop_comment_errors is not None:
                    return None, Error(
                        prop.description.parsed.node,
                        f"Failed to generate the documentation comment "
                        f"for the property {prop.name!r}",
                        prop_comment_errors,
                    )

                blocks.append(
                    Stripped(
                        f"{prop_comment}\n"
                        f"public {prop_type} {prop_name} {{ get; set; }}"
                    )
                )
            else:
                blocks.append(
                    Stripped(f"public {prop_type} {prop_name} {{ get; set; }}")
                )

    # endregion

    # region Signatures

    for signature in interface.signatures:
        signature_blocks = []  # type: List[Stripped]

        if signature.description is not None:
            (
                signature_comment,
                signature_comment_errors,
            ) = csharp_description.generate_signature_comment(signature.description)

            if signature_comment_errors is not None:
                return None, Error(
                    signature.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for signature {signature.name!r}",
                    signature_comment_errors,
                )

            assert signature_comment is not None

            signature_blocks.append(signature_comment)

        # fmt: off
        returns = (
            csharp_common.generate_type(type_annotation=signature.returns)
            if signature.returns is not None else "void"
        )
        # fmt: on

        arg_codes = []  # type: List[Stripped]
        for arg in signature.arguments:
            arg_type = csharp_common.generate_type(type_annotation=arg.type_annotation)
            arg_name = csharp_naming.argument_name(arg.name)
            arg_codes.append(Stripped(f"{arg_type} {arg_name}"))

        signature_name = csharp_naming.method_name(signature.name)
        if len(arg_codes) > 2:
            arg_block = ",\n".join(arg_codes)
            arg_block_indented = textwrap.indent(arg_block, I)
            signature_blocks.append(
                Stripped(f"public {returns} {signature_name}(\n{arg_block_indented});")
            )
        elif len(arg_codes) == 1:
            signature_blocks.append(
                Stripped(f"public {returns} {signature_name}({arg_codes[0]});")
            )
        else:
            assert len(arg_codes) == 0
            signature_blocks.append(Stripped(f"public {returns} {signature_name}();"))

        blocks.append(Stripped("\n".join(signature_blocks)))

    # endregion

    for i, code in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(code, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


class _DescendBodyUnroller(csharp_unrolling.Unroller):
    """Generate the code that unrolls descent into an element."""

    #: If set, generates the code with unrolled yields.
    #: Otherwise, we do not unroll recursively.
    _recurse: Final[bool]

    #: Pre-computed descendability map. A type is descendable if we should unroll it
    #: further.
    _descendability: Final[Mapping[intermediate.TypeAnnotationUnion, bool]]

    def __init__(
        self,
        recurse: bool,
        descendability: Mapping[intermediate.TypeAnnotationUnion, bool],
    ) -> None:
        """Initialize with the given values."""
        self._recurse = recurse
        self._descendability = descendability

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        # We can not descend into a primitive type.
        return []

    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        symbol = type_annotation.symbol

        if isinstance(symbol, intermediate.Enumeration):
            return []

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # We can not descend into a primitive type.
            return []

        assert isinstance(symbol, intermediate.Class)  # Exhaustively match

        result = [csharp_unrolling.Node(f"yield return {unrollee_expr};", children=[])]

        if self._recurse:
            if self._descendability[type_annotation]:
                recurse_var = csharp_unrolling.Unroller._loop_var_name(
                    level=item_level, suffix="Item"
                )

                result.append(
                    csharp_unrolling.Node(
                        text=textwrap.dedent(
                            f"""\
                        // Recurse
                        foreach (var {recurse_var} in {unrollee_expr}.Descend())
                        {{
                            yield return {recurse_var};
                        }}"""
                        ),
                        children=[],
                    )
                )
            else:
                result.append(
                    csharp_unrolling.Node(
                        text="// Recursive descent ends here.", children=[]
                    )
                )

        return result

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        item_var = csharp_unrolling.Unroller._loop_var_name(
            level=item_level, suffix="Item"
        )

        children = self.unroll(
            unrollee_expr=item_var,
            type_annotation=type_annotation.items,
            path=[],  # Path is unused in this context
            item_level=item_level + 1,
            key_value_level=key_value_level,
        )

        if len(children) == 0:
            return []

        node = csharp_unrolling.Node(
            text=f"foreach (var {item_var} in {unrollee_expr})",
            children=children,
        )

        return [node]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )

        if len(children) == 0:
            return []

        return [
            csharp_unrolling.Node(
                text=f"if ({unrollee_expr} != null)", children=children
            )
        ]


def _generate_descend_body(cls: intermediate.ConcreteClass, recurse: bool) -> Stripped:
    """
    Generate the body of the ``Descend`` and ``DescendOnce`` methods.

    With this function, we can unroll the recursion as a simple optimization
    in the recursive case.
    """
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        descendability = intermediate.map_descendability(
            type_annotation=prop.type_annotation
        )

        if not descendability[prop.type_annotation]:
            continue

        # region Unroll

        unroller = _DescendBodyUnroller(recurse=recurse, descendability=descendability)

        roots = unroller.unroll(
            unrollee_expr=csharp_naming.property_name(prop.name),
            type_annotation=prop.type_annotation,
            path=[],  # We do not use path in this context
            item_level=0,
            key_value_level=0,
        )

        assert len(roots) > 0, (
            "Since the type annotation was descendable, we must have obtained "
            "at least one unrolling node"
        )

        blocks.extend(Stripped(csharp_unrolling.render(root)) for root in roots)

        # endregion

    if len(blocks) == 0:
        blocks.append(Stripped("// No descendable properties\nyield break;"))

    return Stripped("\n\n".join(blocks))


def _generate_descend_once_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the ``DescendOnce`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=False)

    indented_body = textwrap.indent(body, I)

    return Stripped(
        f"""\
/// <summary>
/// Iterate over all the class instances referenced from this instance
/// without further recursion.
/// </summary>
public IEnumerable<IClass> DescendOnce()
{{
{indented_body}
}}"""
    )


def _generate_descend_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the recursive ``Descend`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=True)

    indented_body = textwrap.indent(body, I)

    return Stripped(
        f"""\
/// <summary>
/// Iterate recursively over all the class instances referenced from this instance.
/// </summary>
public IEnumerable<IClass> Descend()
{{
{indented_body}
}}"""
    )


def _generate_default_value(default: intermediate.Default) -> Stripped:
    """Generate the C# code representing the default value of an argument."""
    code = None  # type: Optional[str]

    if default is not None:
        if isinstance(default, intermediate.DefaultConstant):
            if default.value is None:
                code = "null"
            elif isinstance(default.value, bool):
                code = "true" if default.value else "false"
            elif isinstance(default.value, str):
                code = csharp_common.string_literal(default.value)
            elif isinstance(default.value, int):
                code = str(default.value)
            elif isinstance(default.value, float):
                code = f"{default}d"
            else:
                assert_never(default.value)
        elif isinstance(default, intermediate.DefaultEnumerationLiteral):
            code = ".".join(
                [
                    csharp_naming.enum_name(default.enumeration.name),
                    csharp_naming.enum_literal_name(default.literal.name),
                ]
            )
        else:
            assert_never(default)

    assert code is not None
    return Stripped(code)


@require(lambda cls: not cls.is_implementation_specific)
@require(lambda cls: not cls.constructor.is_implementation_specific)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_constructor(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the constructor function for the given concrete class ``cls``."""
    cls_name = csharp_naming.class_name(cls.name)

    blocks = []  # type: List[str]

    arg_codes = []  # type: List[str]
    for arg in cls.constructor.arguments:
        arg_type = csharp_common.generate_type(type_annotation=arg.type_annotation)
        arg_name = csharp_naming.argument_name(arg.name)

        if arg.default is None:
            arg_codes.append(Stripped(f"{arg_type} {arg_name}"))
        else:
            arg_codes.append(
                Stripped(
                    f"{arg_type} {arg_name} = {_generate_default_value(arg.default)}"
                )
            )

    if len(arg_codes) == 0:
        return (
            None,
            Error(
                cls.parsed.node,
                "An empty constructor is automatically generated, "
                "which conflicts with the empty constructor "
                "specified in the meta-model",
            ),
        )

    elif len(arg_codes) == 1:
        blocks.append(f"public {cls_name}({arg_codes[0]})\n{{")
    else:
        arg_block = ",\n".join(arg_codes)
        arg_block_indented = textwrap.indent(arg_block, I)
        blocks.append(Stripped(f"public {cls_name}(\n{arg_block_indented})\n{{"))

    body = []  # type: List[str]
    for stmt in cls.constructor.statements:
        if isinstance(stmt, intermediate_construction.AssignArgument):
            if stmt.default is None:
                body.append(
                    f"{csharp_naming.property_name(stmt.name)} = "
                    f"{csharp_naming.argument_name(stmt.argument)};"
                )
            else:
                if isinstance(stmt.default, intermediate_construction.EmptyList):
                    prop = cls.properties_by_name[stmt.name]

                    type_anno = prop.type_annotation
                    while isinstance(type_anno, intermediate.OptionalTypeAnnotation):
                        type_anno = type_anno.value

                    prop_type = csharp_common.generate_type(type_annotation=type_anno)

                    arg_name = csharp_naming.argument_name(stmt.argument)

                    # Write the assignment as a ternary operator
                    writer = io.StringIO()
                    writer.write(f"{csharp_naming.property_name(stmt.name)} = ")
                    writer.write(f"({arg_name} != null)\n")
                    writer.write(textwrap.indent(f"? {arg_name}\n", I))
                    writer.write(textwrap.indent(f": new {prop_type}();", I))

                    body.append(writer.getvalue())
                elif isinstance(
                    stmt.default, intermediate_construction.DefaultEnumLiteral
                ):
                    literal_code = ".".join(
                        [
                            csharp_naming.enum_name(stmt.default.enum.name),
                            csharp_naming.enum_literal_name(stmt.default.literal.name),
                        ]
                    )

                    body.append(
                        f"{csharp_naming.property_name(stmt.name)} = "
                        f"{literal_code};"
                    )
                else:
                    assert_never(stmt.default)

        else:
            assert_never(stmt)

    blocks.append("\n".join(textwrap.indent(stmt_code, I) for stmt_code in body))

    blocks.append("}")

    return Stripped("\n".join(blocks)), None


@require(lambda cls: not cls.is_implementation_specific)
@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_class(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate C# code for the given concrete class ``cls``."""
    writer = io.StringIO()

    if cls.description is not None:
        comment, comment_errors = csharp_description.generate_symbol_comment(
            cls.description
        )
        if comment_errors is not None:
            return None, Error(
                cls.description.parsed.node,
                "Failed to generate the comment description",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    name = csharp_naming.class_name(cls.name)

    # NOTE (mristin, 2021-12-15):
    # Since C# does not support multiple inheritance, we model all the abstract classes
    # as interfaces. Hence a class only implements interfaces and does not extend
    # any abstract class.
    #
    # Moreover, if a concrete class has descendants (which we allow in the meta-model),
    # we will additionally generate an interface for that class and that class then also
    # implements the interface which was based on it.
    #
    # Finally, every class of the meta-model also implements the general
    # ``IClass`` interface.

    interface_names = []  # type: List[Identifier]
    for inheritance in cls.inheritances:
        assert inheritance.interface is not None, (
            f"Expected interface in the parent class {inheritance.name!r} "
            f"of class {cls.name!r}"
        )

        interface_names.append(csharp_naming.name_of(inheritance.interface))

    interface_names.append(csharp_naming.interface_name(Identifier("Class")))

    assert len(interface_names) > 0
    if len(interface_names) == 1:
        writer.write(f"public class {name} : {interface_names[0]}\n{{\n")
    else:
        writer.write(f"public class {name} :\n")
        for i, interface_name in enumerate(interface_names):
            if i > 0:
                writer.write(",\n")

            writer.write(textwrap.indent(interface_name, II))

        writer.write("\n{\n")

    # Code blocks separated by double newlines and indented once
    blocks = []  # type: List[Stripped]

    # region Getters and setters

    for prop in cls.properties:
        prop_type = csharp_common.generate_type(type_annotation=prop.type_annotation)

        prop_name = csharp_naming.property_name(prop.name)

        prop_blocks = []  # type: List[Stripped]

        if prop.description is not None:
            (
                prop_comment,
                prop_comment_errors,
            ) = csharp_description.generate_property_comment(prop.description)
            if prop_comment_errors:
                return None, Error(
                    prop.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for the property {prop.name!r}",
                    prop_comment_errors,
                )

            assert prop_comment is not None

            prop_blocks.append(prop_comment)

        prop_blocks.append(Stripped(f"public {prop_type} {prop_name} {{ get; set; }}"))

        blocks.append(Stripped("\n".join(prop_blocks)))

    # endregion

    # region Methods

    errors = []  # type: List[Error]

    for method in cls.methods:
        if isinstance(method, intermediate.ImplementationSpecificMethod):
            # NOTE (mristin, 2022-05-18):
            # We have to repeat the implementation of the method in all the descendants
            # since we share only interfaces between the classes, but not
            # the implementations.
            #
            # This makes the code a bit larger, but the class hierarchy is much simpler
            # and the individual classes are much easier to grasp.
            implementation_key = specific_implementations.ImplementationKey(
                f"Types/{method.specified_for.name}/{method.name}.cs"
            )

            implementation = spec_impls.get(implementation_key, None)

            if implementation is None:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"The implementation is missing for "
                        f"the implementation-specific method: {implementation_key}",
                    )
                )
                continue

            blocks.append(implementation)
        else:
            # NOTE (mristin, 2021-09-16):
            # At the moment, we do not transpile the method body and its contracts.
            # We want to finish the meta-model for the V3 and fix de/serialization
            # before taking on this rather hard task.

            errors.append(
                Error(
                    cls.parsed.node,
                    "At the moment, we do not transpile the method body and "
                    "its contracts. We want to finish the meta-model for the V3 and "
                    "fix de/serialization before taking on this rather hard task.",
                )
            )

    blocks.append(_generate_descend_once_method(cls=cls))

    blocks.append(_generate_descend_method(cls=cls))

    blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Accept the <paramref name="visitor" /> to visit this instance
/// for double dispatch.
/// </summary>
public void Accept(Visitation.IVisitor visitor)
{{
{I}visitor.Visit(this);
}}"""
        )
    )

    blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Accept the visitor to visit this instance for double dispatch
/// with the <paramref name="context" />.
/// </summary>
public void Accept<C>(Visitation.IVisitorWithContext<C> visitor, C context)
{{
{I}visitor.Visit(this, context);
}}"""
        )
    )

    blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Accept the <paramref name="transformer" /> to transform this instance
/// for double dispatch.
/// </summary>
public T Transform<T>(Visitation.ITransformer<T> transformer)
{{
{I}return transformer.Transform(this);
}}"""
        )
    )

    blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Accept the <paramref name="transformer" /> to visit this instance
/// for double dispatch with the <paramref name="context" />.
/// </summary>
public T Transform<C, T>(
{I}Visitation.ITransformerWithContext<C, T> transformer, C context)
{{
{I}return transformer.Transform(this, context);
}}"""
        )
    )

    # endregion

    # region Constructor

    if cls.constructor.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"Types/{cls.name}/{cls.name}.cs"
        )
        implementation = spec_impls.get(implementation_key, None)

        if implementation is None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"The implementation of the implementation-specific constructor "
                    f"is missing: {implementation_key}",
                )
            )
        else:
            blocks.append(implementation)
    else:
        constructor_block, error = _generate_constructor(cls=cls)

        if error is not None:
            errors.append(error)
        else:
            assert constructor_block is not None
            blocks.append(constructor_block)

    # endregion

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the code for the class {cls.name}",
            errors,
        )

    for i, code in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(code, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: VerifiedIntermediateSymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code of the structures based on the symbol table.

    The ``namespace`` defines the AAS C# namespace.
    """
    blocks = [csharp_common.WARNING]  # type: List[Rstripped]

    using_directives = [
        "using EnumMemberAttribute = System.Runtime.Serialization.EnumMemberAttribute;",
        "using System.Collections.Generic;  // can't alias",
    ]  # type: List[str]

    if len(using_directives) > 0:
        blocks.append(Stripped("\n".join(using_directives)))

    blocks.append(Stripped(f"namespace {namespace}\n{{"))

    blocks.append(
        Rstripped(
            textwrap.indent(
                f"""\
/// <summary>
/// Represent a general class of an AAS model.
/// </summary>
public interface IClass
{{
{I}/// <summary>
{I}/// Iterate over all the class instances referenced from this instance
{I}/// without further recursion.
{I}/// </summary>
{I}public IEnumerable<IClass> DescendOnce();

{I}/// <summary>
{I}/// Iterate recursively over all the class instances referenced from this instance.
{I}/// </summary>
{I}public IEnumerable<IClass> Descend();

{I}/// <summary>
{I}/// Accept the <paramref name="visitor" /> to visit this instance
{I}/// for double dispatch.
{I}/// </summary>
{I}public void Accept(Visitation.IVisitor visitor);

{I}/// <summary>
{I}/// Accept the visitor to visit this instance for double dispatch
{I}/// with the <paramref name="context" />.
{I}/// </summary>
{I}public void Accept<C>(Visitation.IVisitorWithContext<C> visitor, C context);

{I}/// <summary>
{I}/// Accept the <paramref name="transformer" /> to transform this instance
{I}/// for double dispatch.
{I}/// </summary>
{I}public T Transform<T>(Visitation.ITransformer<T> transformer);

{I}/// <summary>
{I}/// Accept the <paramref name="transformer" /> to visit this instance
{I}/// for double dispatch with the <paramref name="context" />.
{I}/// </summary>
{I}public T Transform<C, T>(
{II}Visitation.ITransformerWithContext<C, T> transformer, C context);
}}""",
                I,
            )
        )
    )

    errors = []  # type: List[Error]

    for something in csharp_common.over_enumerations_classes_and_interfaces(
        symbol_table
    ):
        code = None  # type: Optional[Stripped]
        error = None  # type: Optional[Error]

        if (
            isinstance(something, intermediate.Class)
            and something.is_implementation_specific
        ):
            implementation_key = specific_implementations.ImplementationKey(
                f"Types/{something.name}.cs"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                error = Error(
                    something.parsed.node,
                    f"The implementation is missing "
                    f"for the implementation-specific class: {implementation_key}",
                )
        else:
            if isinstance(something, intermediate.Enumeration):
                # BEFORE-RELEASE (mristin, 2021-12-13): test in isolation
                code, error = _generate_enum(enum=something)
            elif isinstance(something, intermediate.Interface):
                # BEFORE-RELEASE (mristin, 2021-12-13): test in isolation
                code, error = _generate_interface(interface=something)

            elif isinstance(something, intermediate.ConcreteClass):
                # BEFORE-RELEASE (mristin, 2021-12-13): test in isolation
                code, error = _generate_class(cls=something, spec_impls=spec_impls)
            else:
                assert_never(something)

        assert (code is None) ^ (error is None)
        if error is not None:
            errors.append(
                Error(
                    something.parsed.node,
                    f"Failed to generate the code for {something.name!r}",
                    [error],
                )
            )
        else:
            assert code is not None
            blocks.append(Rstripped(textwrap.indent(code, "    ")))

    if len(errors) > 0:
        return None, errors

    blocks.append(Rstripped(f"}}  // namespace {namespace}"))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Provide data structures for the constraint inferences."""
from typing import Mapping, Sequence, Optional

from icontract import require

from aas_core_codegen import intermediate


class LenConstraint:
    """
    Represent the inferred constraint on the ``len`` of something.

    Both bounds are inclusive: ``min_value ≤ len ≤ max_value``.
    """

    # fmt: off
    @require(
        lambda min_value, max_value:
        not (min_value is not None and max_value is not None)
        or 0 < min_value <= max_value
    )
    # fmt: on
    def __init__(self, min_value: Optional[int], max_value: Optional[int]) -> None:
        """Initialize with the given values."""
        self.min_value = min_value
        self.max_value = max_value

    def copy(self) -> "LenConstraint":
        """Create a copy of the self."""
        return LenConstraint(min_value=self.min_value, max_value=self.max_value)

    def __str__(self) -> str:
        return (
            f"LenConstraint(min_value={self.min_value!r}, max_value={self.max_value!r})"
        )


class PatternConstraint:
    """Constrain a string to comply to a regular expression."""

    def __init__(self, pattern: str) -> None:
        """Initialize with the given values."""
        self.pattern = pattern


class ConstraintsByProperty:
    """
    Represent all the inferred property constraints of a symbol.

    The constraints coming from the constrained primitives are in-lined and hence also
    included in this representation.
    """

    def __init__(
        self,
        len_constraints_by_property: Mapping[intermediate.Property, LenConstraint],
        patterns_by_property: Mapping[
            intermediate.Property, Sequence[PatternConstraint]
        ],
    ) -> None:
        """Initialize with the given values."""
        self.len_constraints_by_property = len_constraints_by_property
        self.patterns_by_property = patterns_by_property



"""
Provide common functions for different algorithms for inference of the constraints.

The constraints are inferred based on the invariants, so the properties in this context
refer to member access in form of ``self.some_property``.
"""
from typing import Optional, Union

from aas_core_codegen.common import Identifier

from aas_core_codegen.parse import tree as parse_tree


def match_property(node: parse_tree.Node) -> Optional[Identifier]:
    """
    Match an access to a property.

    For example, matching ``self.something`` will be a ``"something"``.
    """
    if (
        isinstance(node, parse_tree.Member)
        and isinstance(node.instance, parse_tree.Name)
        and node.instance.identifier == "self"
    ):
        return node.name

    return None


class SingleArgFunctionOnMemberOrName:
    """
    Match a call of a function with a single argument on a member or a variable.

    Mind that we only match arguments, but ignore keyword arguments. This is due to
    simplicity of the implementation and this behavior might change in the future.
    """

    def __init__(
        self,
        function_name: Identifier,
        member_or_name: Union[parse_tree.Member, parse_tree.Name],
    ) -> None:
        """Initialize with the given values."""
        self.function_name = function_name
        self.member_or_name = member_or_name


def match_single_arg_function_on_member_or_name(
    node: parse_tree.Node,
) -> Optional[SingleArgFunctionOnMemberOrName]:
    """
    Match a call of a function with a single argument on a member or a variable.

    For example, something like ``len(self.some_property)`` or ``is_MIME_type(self)``.

    Mind that we only match arguments, but ignore keyword arguments. This is due to
    simplicity of the implementation and this behavior might change in the future.
    """
    if not isinstance(node, parse_tree.FunctionCall):
        return None

    if len(node.args) != 1:
        return None

    if not isinstance(node.args[0], (parse_tree.Name, parse_tree.Member)):
        return None

    # noinspection PyTypeChecker
    return SingleArgFunctionOnMemberOrName(
        function_name=node.name.identifier, member_or_name=node.args[0]
    )


class ConditionalOnProp:
    """Represent an invariant conditioned on an optional property."""

    def __init__(
        self, prop_name: Identifier, consequent: parse_tree.Expression
    ) -> None:
        """Initialize with the given values."""
        self.prop_name = prop_name
        self.consequent = consequent


def match_conditional_on_prop(node: parse_tree.Node) -> Optional[ConditionalOnProp]:
    """
    Match an invariant conditioned on an optional property.

    For example, ``not (self.some_prop is not None) or ...`` or
    ``self.some_prop is None or ...``.
    """
    if isinstance(node, parse_tree.Implication):
        if not isinstance(node.antecedent, parse_tree.IsNotNone):
            return None

        prop_name = match_property(node.antecedent.value)
        if prop_name is None:
            return None

        return ConditionalOnProp(prop_name=prop_name, consequent=node.consequent)

    elif isinstance(node, parse_tree.Or):
        if len(node.values) != 2:
            return None

        if not isinstance(node.values[0], parse_tree.IsNone):
            return None

        # noinspection PyUnresolvedReferences
        prop_name = match_property(node.values[0].value)
        if prop_name is None:
            return None

        return ConditionalOnProp(prop_name=prop_name, consequent=node.values[1])

    else:
        return None



"""Infer the constraints on a property based on regular expressions."""
from typing import (
    Optional,
    List,
    MutableMapping,
    Sequence,
    Mapping,
    cast,
    Iterator,
)

from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier
from aas_core_codegen.infer_for_schema import _common as infer_for_schema_common
from aas_core_codegen.infer_for_schema._types import PatternConstraint
from aas_core_codegen.parse import tree as parse_tree


class PatternVerificationsByName(Mapping[Identifier, intermediate.PatternVerification]):
    """Map verification functions based on patterns by their names."""

    def __getitem__(self, k: Identifier) -> intermediate.PatternVerification:
        # Only for type annotation
        raise NotImplementedError()

    def __len__(self) -> int:
        # Only for type annotation
        raise NotImplementedError()

    def __iter__(self) -> Iterator[Identifier]:
        # Only for type annotation
        raise NotImplementedError()

    # fmt: off
    # noinspection PyAbstractClass
    @require(
        lambda mapping:
        all(
            name == func.name
            for name, func in mapping.items()
        )
    )
    # fmt: on
    def __new__(
        cls, mapping: Mapping[Identifier, intermediate.PatternVerification]
    ) -> "PatternVerificationsByName":
        return cast(PatternVerificationsByName, mapping)


def map_pattern_verifications_by_name(
    verifications: Sequence[intermediate.Verification],
) -> PatternVerificationsByName:
    """
    Go over all verifications and map the pattern verifications by their name.

    The verifications which do not perform pattern matching are ignored.
    """
    result = (
        dict()
    )  # type: MutableMapping[Identifier, intermediate.PatternVerification]

    for verification in verifications:
        if isinstance(verification, intermediate.PatternVerification):
            result[verification.name] = verification

    return PatternVerificationsByName(result)


class _ConstraintOnProperty:
    """Represent a match on an expression like ``is_ID_short(self.something)``."""

    def __init__(self, prop_name: Identifier, constraint: PatternConstraint) -> None:
        """Initialize with the given values."""
        self.prop_name = prop_name
        self.constraint = constraint


def _match_constraint_on_property(
    node: parse_tree.Node, pattern_verifications_by_name: PatternVerificationsByName
) -> Optional[_ConstraintOnProperty]:
    """
    Match the pattern constraints on a property.

    Return the match, or None if not matched.
    """
    if not isinstance(node, parse_tree.FunctionCall):
        return None

    if len(node.args) != 1:
        return None

    prop_name = infer_for_schema_common.match_property(node.args[0])
    if prop_name is None:
        return None

    pattern_verification = pattern_verifications_by_name.get(node.name.identifier, None)
    if pattern_verification is None:
        return None

    return _ConstraintOnProperty(
        prop_name=prop_name,
        constraint=PatternConstraint(pattern=pattern_verification.pattern),
    )


def patterns_from_invariants(
    cls: intermediate.Class,
    pattern_verifications_by_name: PatternVerificationsByName,
) -> MutableMapping[intermediate.Property, List[PatternConstraint]]:
    """
    Infer the pattern constraints for every property of the class ``cls``.

    Even if a property is optional, the constraint will still be inferred, since
    schemas usually separate constraints from defining optional/required fields
    (*e.g.*, see JSON schema).

    The constraints are not exhaustive. We only infer constraints based on invariants
    which involve pre-defined list of functions. It might be that the actual invariants
    are tighter.

    The list of patterns means that *all* the patterns need to be satisfied, *i.e.*
    there is a conjunction of patterns.
    """
    # NOTE (mristin, 2021-11-30):
    # We iterate only once through the invariants instead of inferring the constraints
    # for each property individually to be able to keep linear time complexity.

    constraints_on_props = []  # type: List[_ConstraintOnProperty]

    for invariant in cls.invariants:
        # NOTE (mristin, 2022-01-02):
        # We consider only the genuine invariants of the class, and ignore
        # the invariants of its ancestors.
        if invariant.specified_for is not cls:
            continue

        # Match something like ``self.something is None or is_ID_short(self.something)``
        conditional_on_prop = infer_for_schema_common.match_conditional_on_prop(
            invariant.body
        )

        if conditional_on_prop is not None:
            # Match something like
            # ``is_ID_short(self.something) and is_MIME(self.something)``
            if isinstance(conditional_on_prop.consequent, parse_tree.And):
                for value_node in conditional_on_prop.consequent.values:
                    constraint_on_prop = _match_constraint_on_property(
                        node=value_node,
                        pattern_verifications_by_name=pattern_verifications_by_name,
                    )

                    if (
                        constraint_on_prop is not None
                        and constraint_on_prop.prop_name
                        == conditional_on_prop.prop_name
                    ):
                        constraints_on_props.append(constraint_on_prop)

            # Match something like ``is_ID_short(self.something)``
            elif isinstance(conditional_on_prop.consequent, parse_tree.FunctionCall):
                constraint_on_prop = _match_constraint_on_property(
                    node=conditional_on_prop.consequent,
                    pattern_verifications_by_name=pattern_verifications_by_name,
                )

                if constraint_on_prop is not None:
                    constraints_on_props.append(constraint_on_prop)

            else:
                # No match
                pass

        else:
            # Match something like
            # ``is_ID_short(self.something) and is_MIME(self.something)``
            if isinstance(invariant.body, parse_tree.And):
                for value_node in invariant.body.values:
                    constraint_on_prop = _match_constraint_on_property(
                        node=value_node,
                        pattern_verifications_by_name=pattern_verifications_by_name,
                    )

                    if constraint_on_prop is not None:
                        constraints_on_props.append(constraint_on_prop)

            # Match something like ``is_ID_short(self.something)``
            elif isinstance(invariant.body, parse_tree.FunctionCall):
                constraint_on_prop = _match_constraint_on_property(
                    node=invariant.body,
                    pattern_verifications_by_name=pattern_verifications_by_name,
                )

                if constraint_on_prop is not None:
                    constraints_on_props.append(constraint_on_prop)

            else:
                # No match
                pass

    # endregion

    # region Group the constraints by property

    result = (
        dict()
    )  # type: MutableMapping[intermediate.Property, List[PatternConstraint]]

    for constraint_on_prop in constraints_on_props:
        prop = cls.properties_by_name.get(constraint_on_prop.prop_name, None)
        if prop is None:
            continue

        constraints_for_prop = result.get(prop, None)
        if constraints_for_prop is None:
            constraints_for_prop = []
            result[prop] = constraints_for_prop

        constraints_for_prop.append(constraint_on_prop.constraint)

    # endregion

    return result


def _match_pattern_on_self(
    node: parse_tree.Node, pattern_verifications_by_name: PatternVerificationsByName
) -> Optional[PatternConstraint]:
    """
    Match the pattern constraints on ``self``.

    Return the match, or None if not matched.
    """
    if not isinstance(node, parse_tree.FunctionCall):
        return None

    if len(node.args) != 1:
        return None

    if not isinstance(node.args[0], parse_tree.Name):
        return None

    # noinspection PyUnresolvedReferences
    if node.args[0].identifier != "self":
        return None

    pattern_verification = pattern_verifications_by_name.get(node.name.identifier, None)
    if pattern_verification is None:
        return None

    return PatternConstraint(pattern=pattern_verification.pattern)


# fmt: off
@require(
    lambda constrained_primitive:
    constrained_primitive.constrainee == intermediate.PrimitiveType.STR,
    "Infer only patterns on constrained strings"
)
# fmt: on
def infer_patterns_on_self(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    pattern_verifications_by_name: PatternVerificationsByName,
) -> List[PatternConstraint]:
    """
    Infer the pattern constraints for the constrained string ``constrained_primitive``.

    The constraints are not exhaustive. We only infer constraints based on invariants
    which involve pre-defined list of functions. It might be that the actual invariants
    are tighter.

    The list of patterns means that *all* the patterns need to be satisfied, *i.e.*
    there is a conjunction of patterns.
    """
    result = []  # type: List[PatternConstraint]

    for invariant in constrained_primitive.invariants:
        # NOTE (mristin, 2022-01-02):
        # We consider only the genuine invariants of the constrained primitive, and
        # ignore the invariants of its ancestors.
        if invariant.specified_for is not constrained_primitive:
            continue

        # Match something like
        # ``is_ID_short(self) and is_MIME(self)``
        if isinstance(invariant.body, parse_tree.And):
            for value_node in invariant.body.values:
                pattern_on_self = _match_pattern_on_self(
                    node=value_node,
                    pattern_verifications_by_name=pattern_verifications_by_name,
                )

                if pattern_on_self is not None:
                    result.append(pattern_on_self)

        # Match something like ``is_ID_short(self)``
        elif isinstance(invariant.body, parse_tree.FunctionCall):
            pattern_on_self = _match_pattern_on_self(
                node=invariant.body,
                pattern_verifications_by_name=pattern_verifications_by_name,
            )

            if pattern_on_self is not None:
                result.append(pattern_on_self)

        else:
            # No match
            pass

    return result



# pylint: disable=missing-docstring

import contextlib
import io
import os
import pathlib
import tempfile
import unittest

import aas_core_meta.v3rc2

import aas_core_codegen.main


class Test_against_recorded(unittest.TestCase):
    RERECORD = os.environ.get("AAS_CORE_CODEGEN_RERECORD", "").lower() in (
        "1",
        "true",
        "on",
    )

    def test_cases(self) -> None:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

        parent_case_dir = repo_dir / "test_data" / "csharp" / "test_main"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        for module in [aas_core_meta.v3rc2]:
            case_dir = parent_case_dir / module.__name__
            assert case_dir.is_dir(), case_dir

            assert (
                module.__file__ is not None
            ), f"Expected the module {module!r} to have a __file__, but it has None"
            model_pth = pathlib.Path(module.__file__)
            assert model_pth.exists() and model_pth.is_file(), model_pth

            snippets_dir = case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            expected_output_dir = case_dir / "expected_output"

            with contextlib.ExitStack() as exit_stack:
                if Test_against_recorded.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    tmp_dir = (
                        tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
                    )
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = aas_core_codegen.main.Parameters(
                    model_path=model_pth,
                    target=aas_core_codegen.main.Target.CSHARP,
                    snippets_dir=snippets_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()

                return_code = aas_core_codegen.main.execute(
                    params=params, stdout=stdout, stderr=stderr
                )

                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                stdout_pth = expected_output_dir / "stdout.txt"
                normalized_stdout = stdout.getvalue().replace(
                    str(output_dir), "<output dir>"
                )

                if Test_against_recorded.RERECORD:
                    stdout_pth.write_text(normalized_stdout, encoding="utf-8")
                else:
                    self.assertEqual(
                        normalized_stdout,
                        stdout_pth.read_text(encoding="utf-8"),
                        stdout_pth,
                    )

                # BEFORE-RELEASE (mristin, 2021-12-13):
                #  check the remainder of the generated files
                for relevant_rel_pth in [
                    pathlib.Path("types.cs"),
                    pathlib.Path("visitation.cs"),
                    pathlib.Path("verification.cs"),
                    pathlib.Path("reporting.cs"),
                    pathlib.Path("stringification.cs"),
                    pathlib.Path("jsonization.cs"),
                    pathlib.Path("xmlization.cs"),
                ]:
                    expected_pth = expected_output_dir / relevant_rel_pth
                    output_pth = output_dir / relevant_rel_pth

                    if not output_pth.exists():
                        raise FileNotFoundError(
                            f"The output file is missing: {output_pth}"
                        )

                    try:
                        output = output_pth.read_text(encoding="utf-8")
                    except Exception as exception:
                        raise RuntimeError(
                            f"Failed to read the output from {output_pth}"
                        ) from exception

                    if Test_against_recorded.RERECORD:
                        expected_pth.write_text(output, encoding="utf-8")
                    else:
                        try:
                            expected_output = expected_pth.read_text(encoding="utf-8")
                        except Exception as exception:
                            raise RuntimeError(
                                f"Failed to read the expected output "
                                f"from {expected_pth}"
                            ) from exception

                        self.assertEqual(
                            expected_output,
                            output,
                            f"The files {expected_pth} and {output_pth} do not match.",
                        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import io
import os
import pathlib
import unittest

import aas_core_meta.v3rc1
import aas_core_meta.v3rc2

from aas_core_codegen.smoke import main as smoke_main


class Test_against_recorded(unittest.TestCase):
    RERECORD = os.environ.get("AAS_CORE_CODEGEN_RERECORD", "").lower() in (
        "1",
        "true",
        "on",
    )

    def test_cases(self) -> None:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

        parent_case_dir = repo_dir / "test_data" / "smoke" / "test_main"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        expected_dir = parent_case_dir / "expected"

        model_paths_of_the_expected = sorted(
            list(expected_dir.glob("**/*.py"))
            + [
                pathlib.Path(aas_core_meta.v3rc1.__file__),
                pathlib.Path(aas_core_meta.v3rc2.__file__),
            ]
        )

        for model_pth in model_paths_of_the_expected:
            assert model_pth.is_file(), f"{model_pth=}"

            stderr = io.StringIO()
            return_code = smoke_main.execute(model_path=model_pth, stderr=stderr)
            if return_code != 0:
                parts = [
                    f"Unexpected non-zero return code from the smoke script "
                    f"on {model_pth}: {return_code}"
                ]
                if stderr.getvalue():
                    parts.append(f"The captured STDERR was:\n{stderr.getvalue()}")

                raise AssertionError("\n\n".join(parts))

        unexpected_dir = parent_case_dir / "unexpected"

        for model_pth in sorted(unexpected_dir.glob("**/*.py")):
            stderr = io.StringIO()

            return_code = smoke_main.execute(model_path=model_pth, stderr=stderr)

            if return_code == 0:
                parts = [
                    f"Unexpected zero return code from the smoke script "
                    f"on {model_pth}: {return_code}"
                ]
                if stderr.getvalue():
                    parts.append(f"The captured STDERR was:\n{stderr.getvalue()}")

                raise AssertionError("\n\n".join(parts))

            if stderr.getvalue() == "":
                raise AssertionError(
                    f"Unexpected no STDERR from the smoke script on {model_pth}"
                )

            stderr_pth = model_pth.parent / "expected_stderr.txt"

            normalized_stderr = stderr.getvalue().replace(
                str(model_pth), f"<{model_pth.name}>"
            )

            if Test_against_recorded.RERECORD:
                stderr_pth.write_text(normalized_stderr, encoding="utf-8")
            else:
                self.assertEqual(
                    normalized_stderr,
                    stderr_pth.read_text(encoding="utf-8"),
                    stderr_pth,
                )


if __name__ == "__main__":
    unittest.main()



"""Generate the RDF ontology based on the meta-model."""
import io
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import Stripped, Error, assert_never
from aas_core_codegen.rdf_shacl import (
    naming as rdf_shacl_naming,
    common as rdf_shacl_common,
    _description as rdf_shacl_description,
)
from aas_core_codegen.rdf_shacl.common import INDENT as I, INDENT2 as II


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_summary(
    description: intermediate.DescriptionUnion,
) -> Tuple[Optional[str], Optional[Error]]:
    """Generate the comment text based on the summary in the description."""
    renderer = rdf_shacl_description.Renderer()
    tokens, errors = renderer.transform(description.summary)

    if errors is not None:
        return None, Error(
            description.parsed.node,
            "Failed to generate the description comment",
            [Error(description.parsed.node, message) for message in errors],
        )

    assert tokens is not None

    tokens = rdf_shacl_description.without_redundant_breaks(tokens=tokens)

    parts = []  # type: List[str]
    for token in tokens:
        if isinstance(token, rdf_shacl_description.TokenText):
            parts.append(token.content)
        elif isinstance(token, rdf_shacl_description.TokenLineBreak):
            parts.append("\n")
        elif isinstance(token, rdf_shacl_description.TokenParagraphBreak):
            parts.append("\n\n")
        else:
            assert_never(token)

    result = "".join(parts)

    return result, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_for_enumeration(
    enumeration: intermediate.Enumeration, url_prefix: Stripped
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Define an RDF definition of an enumeration."""
    cls_name = rdf_shacl_naming.class_name(enumeration.name)
    cls_label = rdf_shacl_naming.class_label(enumeration.name)

    writer = io.StringIO()
    writer.write(
        f"""\
###  {url_prefix}/{cls_name}
aas:{cls_name} rdf:type owl:Class ;"""
    )

    writer.write(
        f"\n{I}rdfs:label {rdf_shacl_common.string_literal(cls_label)}^^xsd:string ;"
    )

    errors = []  # type: List[Error]

    if enumeration.description is not None:
        summary, error = _generate_summary(enumeration.description)
        if error is not None:
            errors.append(error)
        else:
            assert summary is not None
            writer.write(
                f"\n{I}rdfs:comment {rdf_shacl_common.string_literal(summary)}@en ;"
            )

    if len(enumeration.literals) > 0:
        writer.write(f"\n{I}owl:oneOf (\n")

        # NOTE (mristin, 2022-04-20):
        # We sort by the literal names so that the URLs are sorted in the resulting
        # schema.
        for literal_name in sorted(
            rdf_shacl_naming.enumeration_literal(literal.name)
            for literal in enumeration.literals
        ):
            writer.write(f"{II}<{url_prefix}/{cls_name}/{literal_name}>\n")

        writer.write(f"{I}) ;")

    writer.write("\n.")

    if len(enumeration.literals) > 0:
        # NOTE (mristin, 2022-04-20):
        # We sort by the literal names so that the URLs are sorted in the resulting
        # schema.
        for literal_name, literal in sorted(
            (rdf_shacl_naming.enumeration_literal(literal.name), literal)
            for literal in enumeration.literals
        ):
            literal_label = rdf_shacl_naming.enumeration_literal_label(literal.name)

            writer.write("\n\n")
            writer.write(
                f"""\
###  {url_prefix}/{cls_name}/{literal_name}
<{url_prefix}/{cls_name}/{literal_name}> rdf:type aas:{cls_name} ;
{I}rdfs:label {rdf_shacl_common.string_literal(literal_label)}^^xsd:string ;"""
            )

            if literal.description is not None:
                summary, error = _generate_summary(literal.description)
                if error is not None:
                    errors.append(error)
                else:
                    assert summary is not None
                    writer.write(
                        f"\n{I}rdfs:comment "
                        f"{rdf_shacl_common.string_literal(summary)}@en ;"
                    )

            writer.write("\n.")

    if len(errors) > 0:
        return None, Error(
            enumeration.parsed.node,
            "Failed to generate the RDF definition",
            underlying=errors,
        )

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_owl_class_for_class(
    cls: intermediate.ClassUnion, url_prefix: Stripped
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the code to define an OWL class."""
    cls_name = rdf_shacl_naming.class_name(cls.name)

    writer = io.StringIO()
    writer.write(f"###  {url_prefix}/{cls_name}\n")
    writer.write(f"aas:{cls_name} rdf:type owl:Class ;\n")

    for inheritance in cls.inheritances:
        writer.write(
            f"{I}rdfs:subClassOf "
            f"aas:{rdf_shacl_naming.class_name(inheritance.name)} ;\n"
        )

    cls_label = rdf_shacl_naming.class_label(cls.name)
    writer.write(
        f"{I}rdfs:label {rdf_shacl_common.string_literal(cls_label)}^^xsd:string ;\n"
    )

    if cls.description is not None:
        summary, error = _generate_summary(cls.description)
        if error is not None:
            return None, error

        assert summary is not None
        writer.write(
            f"{I}rdfs:comment {rdf_shacl_common.string_literal(summary)}@en ;\n"
        )

    writer.write(".")
    return Stripped(writer.getvalue()), None


@require(lambda prop, cls: id(prop) in cls.property_id_set)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_property(
    prop: intermediate.Property,
    cls: intermediate.ClassUnion,
    url_prefix: Stripped,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a property ``prop`` of the intermediate ``symbol``."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    cls_name = rdf_shacl_naming.class_name(cls.name)
    rdfs_domain = f"aas:{cls_name}"

    rdf_type = None  # type: Optional[str]

    missing_implementation = False

    if isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(
            type_anno.symbol,
            (
                intermediate.Enumeration,
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            rdf_type = "owl:ObjectProperty"
        elif isinstance(type_anno.symbol, intermediate.ConstrainedPrimitive):
            rdf_type = "owl:DatatypeProperty"
        else:
            assert_never(type_anno.symbol)

    elif isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        rdf_type = "owl:DatatypeProperty"

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        type_anno_items = intermediate.beneath_optional(type_anno.items)

        if isinstance(type_anno_items, intermediate.OurTypeAnnotation):
            rdf_type = "owl:ObjectProperty"

        elif isinstance(type_anno_items, intermediate.PrimitiveTypeAnnotation):
            rdf_type = "owl:DatatypeProperty"

        else:
            missing_implementation = True
    else:
        missing_implementation = True

    if missing_implementation:
        return None, Error(
            prop.parsed.node,
            f"(mristin, 2021-11-12): "
            f"We did not refine the definition of the non-atomic and non-sequential "
            f"properties. If you see this message, it is time to implement "
            f"this missing functionality.\n"
            f"{prop=}, {type_anno=}, {cls=}",
        )

    assert rdf_type is not None

    prop_name = rdf_shacl_naming.property_name(prop.name)
    prop_label = Stripped(f"has {rdf_shacl_naming.property_label(prop.name)}")
    rdfs_range = rdf_shacl_common.rdfs_range_for_type_annotation(
        type_annotation=type_anno, class_to_rdfs_range=class_to_rdfs_range
    )

    url = f"{url_prefix}/{cls_name}/{prop_name}"
    writer = io.StringIO()
    writer.write(
        f"""\
###  {url}
<{url}> rdf:type {rdf_type} ;
{I}rdfs:label {rdf_shacl_common.string_literal(prop_label)}^^xsd:string ;
{I}rdfs:domain {rdfs_domain} ;
{I}rdfs:range {rdfs_range} ;"""
    )

    if prop.description:
        summary, error = _generate_summary(prop.description)
        if error is not None:
            return None, error

        assert summary is not None

        writer.write(
            f"\n{I}rdfs:comment {rdf_shacl_common.string_literal(summary)}@en ;"
        )

    writer.write("\n.")
    return Stripped(writer.getvalue()), None


# fmt: off
@require(
    lambda cls:
    not isinstance(cls, intermediate.Class)
    or not cls.is_implementation_specific
)
# fmt: on
def _define_for_class(
    cls: intermediate.ClassUnion,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
    url_prefix: Stripped,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition for the intermediate ``symbol``."""
    blocks = []  # type: List[Stripped]
    errors = []  # type: List[Error]

    owl_class, error = _define_owl_class_for_class(cls=cls, url_prefix=url_prefix)
    if error is not None:
        errors.append(error)
    else:
        assert owl_class is not None
        blocks.append(owl_class)

    # NOTE (mristin, 2022-04-20):
    # We sort by the property names so that the URLs appear sorted in the resulting
    # schema.
    for _, prop in sorted(
        (rdf_shacl_naming.property_name(prop.name), prop) for prop in cls.properties
    ):
        if prop.specified_for is not cls:
            continue

        prop_def, error = _define_property(
            prop=prop,
            cls=cls,
            url_prefix=url_prefix,
            class_to_rdfs_range=class_to_rdfs_range,
        )
        if error is not None:
            errors.append(error)
        else:
            assert prop_def is not None
            blocks.append(prop_def)

    if len(errors) > 0:
        return None, Error(
            None, f"Failed to generate the definition for {cls.name}", errors
        )

    return Stripped("\n\n".join(blocks)), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate(
    symbol_table: intermediate.SymbolTable,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
    spec_impls: specific_implementations.SpecificImplementations,
    url_prefix: Stripped,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the RDF ontology based on the ``symbol_table."""
    errors = []  # type: List[Error]

    preamble_key = specific_implementations.ImplementationKey("rdf/preamble.ttl")
    preamble = spec_impls.get(preamble_key, None)
    if preamble is None:
        errors.append(
            Error(
                None,
                f"The implementation snippet for the RDF preamble "
                f"is missing: {preamble_key}",
            )
        )

    if len(errors) > 0:
        return None, errors

    assert preamble is not None
    blocks = [preamble]  # type: List[Stripped]

    for symbol in sorted(
        symbol_table.symbols,
        key=lambda a_symbol: rdf_shacl_naming.class_name(a_symbol.name),
    ):
        if isinstance(symbol, intermediate.Enumeration):
            block, error = _define_for_enumeration(
                enumeration=symbol, url_prefix=url_prefix
            )

            if error is not None:
                errors.append(error)
            else:
                assert block is not None
                blocks.append(block)

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-02-11):
            # We have to in-line the constraints of the constrained primitives directly
            # in SHACL as we do not want to introduce new entities into the ontology.
            pass

        elif isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if symbol.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"rdf/{symbol.name}/owl_class.ttl"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            symbol.parsed.node,
                            f"The implementation snippet for "
                            f"the class {symbol.parsed.name} "
                            f"is missing: {implementation_key}",
                        )
                    )
                else:
                    blocks.append(implementation)

            else:
                block, error = _define_for_class(
                    cls=symbol,
                    class_to_rdfs_range=class_to_rdfs_range,
                    url_prefix=url_prefix,
                )

                if error is not None:
                    errors.append(error)
                else:
                    assert block is not None
                    blocks.append(block)
        else:
            assert_never(symbol)

    if len(errors) > 0:
        return None, errors

    return Stripped("\n\n".join(blocks)), None



"""Generate RDF and SHACL identifiers based on the identifiers from the meta-model."""
from typing import List

from aas_core_codegen.common import Identifier, Stripped
from aas_core_codegen import naming


def class_name(identifier: Identifier) -> Identifier:
    """
    Generate the class name from the intermediate class ``identifier``.

    >>> class_name(Identifier("something"))
    'Something'

    >>> class_name(Identifier("something_to_URL"))
    'SomethingToURL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for part in parts:
        if part in naming.UPPERCASE_ABBREVIATION_SET:
            cased.append(part.upper())
        else:
            cased.append(part.capitalize())

    return Identifier("".join(cased))


_LOWERCASE_WORDS_IN_LABEL = {"to", "in"}


def class_label(identifier: Identifier) -> Stripped:
    """
    Generate the class label from the intermediate class ``identifier``.

    >>> class_label(Identifier("something"))
    'Something'

    >>> class_label(Identifier("something_good_to_URL"))
    'Something Good to URL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for part in parts:
        if part in naming.UPPERCASE_ABBREVIATION_SET:
            cased.append(part.upper())
        elif part in _LOWERCASE_WORDS_IN_LABEL:
            cased.append(part.lower())
        else:
            cased.append(part.capitalize())

    return Stripped(" ".join(cased))


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate a property name based on its meta-model ``identifier``.

    >>> property_name(Identifier("something"))
    'something'

    >>> property_name(Identifier("something_to_URL"))
    'somethingToURL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for i, part in enumerate(parts):
        if i == 0:
            cased.append(part.lower())
        else:
            if part in naming.UPPERCASE_ABBREVIATION_SET:
                cased.append(part.upper())
            else:
                cased.append(part.capitalize())

    return Identifier("".join(cased))


def property_label(identifier: Identifier) -> Stripped:
    """
    Generate the property label from the intermediate class ``identifier``.

    >>> property_label(Identifier("something"))
    'something'

    >>> property_label(Identifier("something_good_to_URL"))
    'something good to URL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for part in parts:
        if part in naming.UPPERCASE_ABBREVIATION_SET:
            cased.append(part.upper())
        else:
            cased.append(part.lower())

    return Stripped(" ".join(cased))


def enumeration_literal(identifier: Identifier) -> Stripped:
    """
    Generate the enumeration literal for its intermediate ``identifier``.

    >>> enumeration_literal(Identifier('something'))
    'Something'

    >>> enumeration_literal(Identifier("something_to_URL"))
    'SomethingToUrl'
    """
    parts = identifier.split("_")
    return Stripped("".join(part.capitalize() for part in parts))


def enumeration_literal_label(identifier: Identifier) -> Stripped:
    """
    Generate the label for an enumeration literal with intermediate ``identifier``.

    >>> enumeration_literal_label(Identifier("something"))
    'Something'

    >>> enumeration_literal_label(Identifier("something_good_to_URL"))
    'Something Good to URL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for part in parts:
        if part in naming.UPPERCASE_ABBREVIATION_SET:
            cased.append(part.upper())
        elif part in _LOWERCASE_WORDS_IN_LABEL:
            cased.append(part.lower())
        else:
            cased.append(part.capitalize())

    return Stripped(" ".join(cased))



"""Generate XML Schema Definition (XSD) corresponding to the meta-model."""

import re
import xml.etree.ElementTree as ET

# noinspection PyUnresolvedReferences
import xml.dom.minidom
from typing import TextIO, MutableMapping, Optional, Tuple, List, Sequence

from icontract import ensure, require

import aas_core_codegen.xsd
from aas_core_codegen import (
    naming,
    specific_implementations,
    intermediate,
    run,
    infer_for_schema,
)
from aas_core_codegen.common import Error, assert_never
from aas_core_codegen.xsd import naming as xsd_naming

assert aas_core_codegen.xsd.__doc__ == __doc__


def _define_for_enumeration(enumeration: intermediate.Enumeration) -> List[ET.Element]:
    """
    Generate the definitions for an ``enumeration``.
    The root element is to be *extended* with the resulting list.
    """
    restriction = ET.Element("xs:restriction", {"base": "xs:string"})
    for literal in enumeration.literals:
        restriction.append(ET.Element("xs:enumeration", {"value": literal.value}))

    element = ET.Element(
        "xs:simpleType", {"name": xsd_naming.type_name(enumeration.name)}
    )
    element.append(restriction)

    return [element]


_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "xs:boolean",
    intermediate.PrimitiveType.INT: "xs:long",
    intermediate.PrimitiveType.FLOAT: "xs:double",
    intermediate.PrimitiveType.STR: "xs:string",
    intermediate.PrimitiveType.BYTEARRAY: "xs:base64Binary",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)

_ESCAPE_BACKSLASH_X_RE = re.compile(r"\\x([a-fA-f0-9]{2})")


def _undo_escaping_backslash_x_in_pattern(pattern: str) -> str:
    """
    Undo the escaping of `\\x??` in the ``pattern``.

    This is necessary since XML Schema Validators do not know how to handle such escape
    sequences in the patterns and need the verbatim characters.
    """
    parts = []  # type: List[str]
    cursor = None  # type: Optional[int]
    for mtch in re.finditer(_ESCAPE_BACKSLASH_X_RE, pattern):
        if cursor is None:
            parts.append(pattern[: mtch.start()])
        else:
            parts.append(pattern[cursor : mtch.start()])

        ascii_code = int(mtch.group(1), base=16)
        character = chr(ascii_code)
        parts.append(character)
        cursor = mtch.end()

    if cursor is None:
        parts.append(pattern)
    else:
        if cursor < len(pattern):
            parts.append(pattern[cursor:])

    return "".join(parts)


def _generate_xs_restriction(
    base_type: intermediate.PrimitiveType,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Optional[ET.Element]:
    """
    Generate the ``xs:restriction`` for the given primitive.

    If there are no length and pattern constraints, return None.
    """
    if len_constraint is None and (
        pattern_constraints is None or (len(pattern_constraints) == 0)
    ):
        return None

    restriction = ET.Element("xs:restriction", {"base": _PRIMITIVE_MAP[base_type]})

    if pattern_constraints is not None and len(pattern_constraints) > 0:
        if len(pattern_constraints) == 1:
            pattern = ET.Element(
                "xs:pattern",
                {
                    "value": _undo_escaping_backslash_x_in_pattern(
                        pattern_constraints[0].pattern
                    )
                },
            )

            restriction.append(pattern)
        else:
            # BEFORE-RELEASE (mristin, 2021-12-13):
            #  test this and check that the XSD makes sense with somebody else!
            parent_restriction = restriction
            for pattern_constraint in pattern_constraints:
                nested_restriction = ET.Element("xs:restriction")
                pattern = ET.Element(
                    "xs:pattern",
                    {
                        "value": _undo_escaping_backslash_x_in_pattern(
                            pattern_constraint.pattern
                        )
                    },
                )

                nested_restriction.append(pattern)
                parent_restriction.append(nested_restriction)
                parent_restriction = nested_restriction

    if len_constraint is not None:
        if len_constraint.min_value is not None:
            min_length = ET.Element(
                "xs:minLength", {"value": str(len_constraint.min_value)}
            )
            restriction.append(min_length)

        if len_constraint.max_value is not None:
            max_length = ET.Element(
                "xs:maxLength", {"value": str(len_constraint.max_value)}
            )
            restriction.append(max_length)

    return restriction


def _generate_xs_element_for_a_primitive_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> ET.Element:
    """
    Generate the ``xs:element`` for a primitive property.

    A primitive property is a property whose type is either a primitive or
    a constrained primitive. The reason why we take these two together is that we
    in-line the constraints for the constrained primitives.

    We do not define the constrained primitives separately in the schema in order to
    avoid the confusion during comparisons between the XSD and the meta-model in
    the book.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # NOTE (mristin, 2022-03-30):
    # Specify the type of the ``type_anno`` here with assert instead of specifying it
    # in the pre-condition to help mypy a bit.
    assert isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and isinstance(type_anno.symbol, intermediate.ConstrainedPrimitive)
    ), f"Expected a primitive or a constrained primitive, but got: {type_anno}"

    base_type = None  # type: Optional[intermediate.PrimitiveType]
    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        base_type = type_anno.a_type
    elif isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.symbol, intermediate.ConstrainedPrimitive
    ):
        base_type = type_anno.symbol.constrainee
    else:
        raise AssertionError(
            f"Unexpected type_anno type {type(type_anno)}: {type_anno}"
        )

    xs_restriction = _generate_xs_restriction(
        base_type=base_type,
        len_constraint=len_constraint,
        pattern_constraints=pattern_constraints,
    )

    xs_element = None  # type: Optional[ET.Element]
    if xs_restriction is None:
        xs_element = ET.Element(
            "xs:element",
            {
                "name": naming.xml_property(prop.name),
                "type": _PRIMITIVE_MAP[base_type],
            },
        )
    else:
        xs_simple_type = ET.Element("xs:simpleType")
        xs_simple_type.append(xs_restriction)

        xs_element = ET.Element("xs:element", {"name": naming.xml_property(prop.name)})
        xs_element.append(xs_simple_type)

    assert xs_element is not None
    return xs_element


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_xs_element_for_a_list_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the ``xs:element`` for a list property."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # NOTE (mristin, 2022-03-30):
    # Specify the ``type_anno`` here with assert instead of specifying it
    # in the pre-condition to help mypy a bit
    assert isinstance(type_anno, intermediate.ListTypeAnnotation)

    min_occurs = "0"
    max_occurs = "unbounded"
    if len_constraint is not None:
        if len_constraint.min_value is not None:
            min_occurs = str(len_constraint.min_value)

        if len_constraint.max_value is not None:
            max_occurs = str(len_constraint.max_value)

    xs_element = None  # type: Optional[ET.Element]

    if isinstance(type_anno.items, intermediate.OurTypeAnnotation):
        # NOTE (mristin, 2021-11-13):
        # We need to nest the elements in the tag element to separate them in the
        # sequence.

        symbol = type_anno.items.symbol

        if isinstance(symbol, intermediate.Enumeration):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of enumerations "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of constrained primitives "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )

        elif isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if isinstance(symbol, intermediate.AbstractClass) or (
                isinstance(symbol, intermediate.ConcreteClass)
                and len(symbol.concrete_descendants) > 0
            ):

                choice_group_name = xsd_naming.choice_group_name(symbol.name)
                xs_group = ET.Element(
                    "xs:group",
                    {
                        "ref": choice_group_name,
                        "minOccurs": min_occurs,
                        "maxOccurs": max_occurs,
                    },
                )

                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(xs_group)

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)
            else:
                assert isinstance(symbol, intermediate.ConcreteClass)
                assert len(symbol.concrete_descendants) == 0

                xs_element_inner = ET.Element(
                    "xs:element",
                    {
                        "name": naming.xml_class_name(symbol.name),
                        "type": xsd_naming.type_name(symbol.name),
                        "minOccurs": min_occurs,
                        "maxOccurs": max_occurs,
                    },
                )
                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(xs_element_inner)

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of constrained primitives "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )
        else:
            assert_never(symbol)
    else:
        return None, Error(
            prop.parsed.node,
            f"We do not know how to specify the list "
            f"for the property {prop.name!r} of {prop.specified_for.name!r} "
            f"in the XSD with the type: {type_anno}",
        )

    assert xs_element is not None
    return xs_element, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_xs_element_for_a_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the definition of an ``xs:element`` for a property."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    xs_element = None  # type: Optional[ET.Element]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        xs_element = _generate_xs_element_for_a_primitive_property(
            prop=prop,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        symbol = type_anno.symbol

        if isinstance(symbol, intermediate.Enumeration):
            xs_element = ET.Element(
                "xs:element",
                {
                    "name": naming.xml_property(prop.name),
                    "type": xsd_naming.type_name(symbol.name),
                },
            )

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            xs_element = _generate_xs_element_for_a_primitive_property(
                prop=prop,
                len_constraint=len_constraint,
                pattern_constraints=pattern_constraints,
            )

        elif isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if isinstance(symbol, intermediate.AbstractClass) or (
                isinstance(symbol, intermediate.ConcreteClass)
                and len(symbol.concrete_descendants) > 0
            ):
                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(
                    ET.Element(
                        "xs:group", {"ref": xsd_naming.choice_group_name(symbol.name)}
                    )
                )

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)
            else:
                assert (
                    isinstance(symbol, intermediate.ConcreteClass)
                    and len(symbol.concrete_descendants) == 0
                )

                xs_element = ET.Element(
                    "xs:element",
                    {
                        "name": naming.xml_property(prop.name),
                        "type": xsd_naming.type_name(symbol.name),
                    },
                )
        else:
            assert_never(type_anno.symbol)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        xs_element, error = _generate_xs_element_for_a_list_property(
            prop=prop, len_constraint=len_constraint
        )
        if error is not None:
            return None, error

        assert xs_element is not None
    else:
        assert_never(type_anno)

    assert xs_element is not None

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        xs_element.attrib["minOccurs"] = "0"

    return xs_element, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_properties(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Define the properties of the ``cls`` as a sequence of tags."""
    sequence = []  # type: List[ET.Element]
    errors = []  # type: List[Error]

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        len_constraint = constraints_by_property.len_constraints_by_property.get(
            prop, None
        )

        pattern_constraints = constraints_by_property.patterns_by_property.get(
            prop, None
        )

        xs_element, error = _generate_xs_element_for_a_property(
            prop=prop,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )
        if error is not None:
            errors.append(error)
        else:
            assert xs_element is not None
            sequence.append(xs_element)

    if len(errors) > 0:
        return None, errors

    return sequence, None


def _generate_xs_group_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the ``xs:group`` representation of the class properties."""
    properties, properties_errors = _define_properties(
        cls=cls, constraints_by_property=constraints_by_property
    )

    if properties_errors is not None:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate xs:group for the class {cls.name!r}",
            properties_errors,
        )

    assert properties is not None

    xs_sequence = ET.Element("xs:sequence")
    for inheritance in cls.inheritances:
        inheritance_xs_group = ET.Element(
            "xs:group", {"ref": xsd_naming.group_name(inheritance.name)}
        )
        xs_sequence.append(inheritance_xs_group)

    xs_sequence.extend(properties)

    xs_group = ET.Element("xs:group", {"name": xsd_naming.group_name(cls.name)})
    xs_group.append(xs_sequence)

    return xs_group, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[List[ET.Element]], Optional[Error]]:
    """
    Generate the definitions for the class ``cls``.

    The root element is to be *extended* with the resulting list.
    """
    # NOTE (mristin, 2022-03-30):
    # We define each set of properties in a group. Then we reference these groups
    # among the complex types.
    # See: https://stackoverflow.com/questions/1198755/xml-schemas-with-multiple-inheritance

    xs_group, xs_group_error = _generate_xs_group_for_class(
        cls=cls, constraints_by_property=constraints_by_property
    )
    if xs_group_error is not None:
        return None, xs_group_error

    assert xs_group is not None

    xs_group_ref = ET.Element("xs:group", {"ref": xsd_naming.group_name(cls.name)})

    xs_sequence = ET.Element("xs:sequence")
    xs_sequence.append(xs_group_ref)

    complex_type = ET.Element(
        "xs:complexType", {"name": xsd_naming.type_name(cls.name)}
    )
    complex_type.append(xs_sequence)

    return [xs_group, complex_type], None


@require(lambda cls: len(cls.concrete_descendants) > 0)
def _generate_choice_group(cls: intermediate.ClassUnion) -> ET.Element:
    """Generate a group that defines a choice of concrete descendants."""
    xs_choice = ET.Element("xs:choice")

    if isinstance(cls, intermediate.ConcreteClass):
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(cls.name),
                    "type": xsd_naming.type_name(cls.name),
                },
            )
        )

    for descendant in cls.concrete_descendants:
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(descendant.name),
                    "type": xsd_naming.type_name(descendant.name),
                },
            )
        )

    xs_group = ET.Element("xs:group", {"name": xsd_naming.choice_group_name(cls.name)})
    xs_group.append(xs_choice)
    return xs_group


_WHITESPACE_RE = re.compile(r"\s+")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _retrieve_implementation_specific_elements(
    cls: intermediate.ClassUnion,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Parse the elements from the implementation-specific snippet."""
    implementation_key = specific_implementations.ImplementationKey(f"{cls.name}.xml")

    text = spec_impls.get(implementation_key, None)
    if text is None:
        return None, [
            Error(
                cls.parsed.node,
                f"The implementation is missing "
                f"for the implementation-specific class: {implementation_key}",
            )
        ]

    # noinspection PyUnusedLocal
    implementation_root = None  # type: Optional[ET.Element]

    try:
        implementation_root = ET.fromstring(text)
    except Exception as err:
        return None, [
            Error(
                cls.parsed.node,
                f"Failed to parse the XML out of "
                f"the specific implementation {implementation_key}: {err}",
            )
        ]

    assert implementation_root is not None

    errors = []  # type: List[Error]
    for descendant in implementation_root.iter():
        if descendant.text is not None and not _WHITESPACE_RE.fullmatch(
            descendant.text
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.text!r}",
                )
            )
            continue

        if descendant.tail is not None and not _WHITESPACE_RE.fullmatch(
            descendant.tail
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected tail text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.tail!r}",
                )
            )
            continue

    if len(errors) > 0:
        return None, errors

    # Ignore the implementation root since it defines a partial schema
    elements = []  # type: List[ET.Element]
    for child in implementation_root:
        elements.append(child)

    return elements, None


def _sort_by_tags_and_names_in_place(root: ET.Element) -> None:
    """
    Sort the children elements by tag and name attribute in place.

    This makes diffing and searching in the schema a bit easier.
    """
    groups = []  # type: List[ET.Element]
    simple_types = []  # type: List[ET.Element]
    complex_types = []  # type: List[ET.Element]
    miscellaneous = []  # type: List[ET.Element]
    elements = []  # type: List[ET.Element]

    for child in root:
        if child.tag == "xs:group":
            groups.append(child)
        elif child.tag == "xs:simpleType":
            simple_types.append(child)
        elif child.tag == "xs:complexType":
            complex_types.append(child)
        elif child.tag == "xs:element":
            elements.append(child)
        else:
            miscellaneous.append(child)

    for element_list in [groups, simple_types, complex_types, miscellaneous, elements]:
        element_list.sort(key=lambda elt: elt.attrib.get("name", ""))

    children = groups + simple_types + complex_types + elements + miscellaneous

    assert len(children) == len(root)
    root[:] = children


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the XML Schema Definition (XSD) based on the ``symbol_table."""
    root_element_key = specific_implementations.ImplementationKey("root_element.xml")

    root_element_as_text = spec_impls.get(root_element_key, None)
    if root_element_as_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing: {root_element_key}",
            )
        ]

    # noinspection PyUnusedLocal
    root = None  # type: Optional[ET.Element]
    try:
        root = ET.fromstring(root_element_as_text)
    except ET.ParseError as err:
        return None, [
            Error(
                None, f"Failed to parse the root element from {root_element_key}: {err}"
            )
        ]

    # NOTE (mristin, 2022-03-30):
    # We need to use minidom to extract the ``xmlns`` property as ElementTree removes
    # it.
    # noinspection PyUnresolvedReferences
    minidom_doc = xml.dom.minidom.parseString(root_element_as_text)
    xmlns = minidom_doc.documentElement.getAttribute("xmlns")

    assert root is not None

    errors = []  # type: List[Error]

    # NOTE (mristin, 2022-04-09):
    # We remove any whitespace tail and text in all the tags, and make sure there is no
    # unexpected text anywhere.
    for element in root.iter():
        if element.text is not None:
            if _WHITESPACE_RE.fullmatch(element.text):
                element.text = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected text in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.text!r}",
                    )
                )

        if element.tail is not None:
            if _WHITESPACE_RE.fullmatch(element.tail):
                element.tail = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected tail in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.tail!r}",
                    )
                )

    if len(errors) > 0:
        return None, errors

    constraints_by_class, some_errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )

    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert constraints_by_class is not None

    ids_of_symbols_in_properties = intermediate.collect_ids_of_symbols_in_properties(
        symbol_table=symbol_table
    )

    for symbol in symbol_table.symbols:
        elements = None  # type: Optional[List[ET.Element]]

        if (
            isinstance(symbol, (intermediate.AbstractClass, intermediate.ConcreteClass))
            and symbol.is_implementation_specific
        ):
            elements, impl_spec_errors = _retrieve_implementation_specific_elements(
                cls=symbol, spec_impls=spec_impls
            )
            if impl_spec_errors is not None:
                errors.extend(impl_spec_errors)
                continue

            assert elements is not None
        else:
            if isinstance(symbol, intermediate.Enumeration):
                if id(symbol) not in ids_of_symbols_in_properties:
                    continue

                elements = _define_for_enumeration(enumeration=symbol)

            elif isinstance(symbol, intermediate.ConstrainedPrimitive):
                # NOTE (mristin, 2022-03-30):
                # We in-line the constraints from the constrained primitives directly
                # in the properties. We do not want to introduce separate definitions
                # for them as that would make it more difficult for downstream code
                # generators to generate meaningful code.

                continue

            elif isinstance(
                symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                elements, definition_error = _define_for_class(
                    cls=symbol, constraints_by_property=constraints_by_class[symbol]
                )

                if definition_error is not None:
                    errors.append(definition_error)
                    continue

                assert elements is not None

                if len(symbol.concrete_descendants) > 0:
                    choice_group = _generate_choice_group(cls=symbol)
                    elements.append(choice_group)
            else:
                assert_never(symbol)

        assert elements is not None
        root.extend(elements)

    if len(errors) > 0:
        return None, errors

    # Tag name 🠒 (name 🠒 element)
    observed_definitions = dict(
        dict()
    )  # type: MutableMapping[str, MutableMapping[str, ET.Element]]

    for element in root:
        name = element.attrib.get("name", None)
        if name is None:
            continue

        observed_for_tag = observed_definitions.get(element.tag, None)
        if observed_for_tag is None:
            observed_for_tag = dict()
            observed_definitions[element.tag] = observed_for_tag

        observed_element = observed_for_tag.get(name, None)
        if observed_element is not None:
            ours = ET.tostring(element, encoding="unicode", method="xml")
            theirs = ET.tostring(observed_element, encoding="unicode", method="xml")

            errors.append(
                Error(
                    None,
                    f"There are conflicting definitions in the schema "
                    f"with the name {name!r}:\n"
                    f"\n"
                    f"{ours}\n"
                    f"\n"
                    f"and\n"
                    f"\n"
                    f"{theirs}",
                )
            )
        else:
            observed_for_tag[name] = element

    if len(errors) > 0:
        return None, errors

    _sort_by_tags_and_names_in_place(root)

    # NOTE (mristin, 2022-03-30):
    # For some unknown reason, ElementTree erases the xmlns property of the root
    # element. Therefore we need to add it here manually.
    root.attrib["xmlns"] = xmlns

    text = ET.tostring(root, encoding="unicode", method="xml")

    # NOTE (mristin, 2021-11-23):
    # This approach is slow, but effective. As long as the meta-model is not too big,
    # this should work.
    # noinspection PyUnresolvedReferences
    pretty_text = xml.dom.minidom.parseString(text).toprettyxml(indent="  ")

    return pretty_text, None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the XML Schema Definition "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "schema.xsd"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the XML Schema Definition to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Run aas-core-codegen as Python module."""

import aas_core_codegen.main

if __name__ == "__main__":
    # The ``prog`` needs to be set in the argparse.
    # Otherwise the program name in the help shown to the user will be ``__main__``.
    aas_core_codegen.main.main(prog="aas_core_codegen")



"""Generate the visitor classes based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

import aas_core_codegen.csharp.common as csharp_common
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II
import aas_core_codegen.csharp.naming as csharp_naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, Stripped, Rstripped, assert_never


# region Generate


def _generate_ivisitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the visitor interface."""
    blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public void Visit({cls_name} that);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Define the interface for a visitor which visits the instances of the model.
/// </summary>
public interface IVisitor
{{
{I}public void Visit(IClass that);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public interface IVisitor")

    return Stripped(writer.getvalue())


def _generate_visitor_through(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the visitor that simply iterates over the instances."""
    blocks = [
        Stripped(
            f"""\
public void Visit(IClass that)
{{
{I}that.Accept(this);
}}"""
        )
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(
                Stripped(
                    f"""\
public void Visit({cls_name} that)
{{
{I}// Just descend through, do nothing with <c>that</c>
{I}foreach (var something in that.DescendOnce())
{I}{{
{II}Visit(something);
{I}}}
}}"""
                )
            )

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Just descend through the instances without any action.
/// </summary>
/// <remarks>
/// This class is meaningless for itself. However, it is a good base if you
/// want to descend through instances and apply actions only on a subset of
/// classes.
/// </remarks>
public class VisitorThrough : IVisitor
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public class VisitorThrough")

    return Stripped(writer.getvalue())


def _generate_abstract_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the visitor that performs double-dispatch."""
    blocks = [
        Stripped(
            f"""\
public void Visit(IClass that)
{{
{I}that.Accept(this);
}}"""
        )
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.Class):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public abstract void Visit({cls_name} that);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Perform double-dispatch to visit the concrete instances.
/// </summary>
public abstract class AbstractVisitor : IVisitor
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public abstract class AbstractVisitor")

    return Stripped(writer.getvalue())


def _generate_ivisitor_with_context(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the interface for the visitor with context."""
    blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public void Visit({cls_name} that, C context);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Define the interface for a visitor which visits the instances of the model.
/// </summary>
/// <typeparam name="C">Context type</typeparam>
public interface IVisitorWithContext<C>
{{
{I}public void Visit(IClass that, C context);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public interface IVisitorWithContext")

    return Stripped(writer.getvalue())


def _generate_abstract_visitor_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the visitor with context that performs double-dispatch."""
    blocks = [
        Stripped(
            f"""\
public void Visit(IClass that, C context)
{{
{I}that.Accept(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not visit them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.Class):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(
                Stripped(f"public abstract void Visit({cls_name} that, C context);")
            )

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Perform double-dispatch to visit the concrete instances
/// with context.
/// </summary>
/// <typeparam name="C">Context type</typeparam>
public abstract class AbstractVisitorWithContext<C>
{I}: IVisitorWithContext<C>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public abstract class AbstractVisitorWithContext")

    return Stripped(writer.getvalue())


def _generate_itransformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the transformer interface."""
    blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not transform them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public T Transform({cls_name} that);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Define the interface for a transformer which transforms recursively
/// the instances into something else.
/// </summary>
/// <typeparam name="T">The type of the transformation result</typeparam>
public interface ITransformer<T>
{{
{I}public T Transform(IClass that);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public interface ITransformer")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the most abstract transformer that merely double-dispatches."""
    blocks = [
        Stripped(
            f"""\
public T Transform(IClass that)
{{
{I}return that.Transform(this);
}}"""
        )
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not transform them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public abstract T Transform({cls_name} that);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Perform double-dispatch to transform recursively
/// the instances into something else.
/// </summary>
/// <typeparam name="T">The type of the transformation result</typeparam>
public abstract class AbstractTransformer<T> : ITransformer<T>
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public abstract class AbstractTransformer")

    return Stripped(writer.getvalue())


def _generate_itransformer_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the interface for the transformer with context."""
    blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not transform them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(Stripped(f"public T Transform({cls_name} that, C context);"))

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Define the interface for a transformer which recursively transforms
/// the instances into something else while the context is passed along.
/// </summary>
/// <typeparam name="C">Type of the transformation context</typeparam>
/// <typeparam name="T">The type of the transformation result</typeparam>
public interface ITransformerWithContext<C, T>
{{
{I}public T Transform(IClass that, C context);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public interface ITransformerWithContext")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the most abstract transformer that merely double-dispatches."""
    blocks = [
        Stripped(
            f"""\
public T Transform(IClass that, C context)
{{
{I}return that.Transform(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # Constrained primitives are modeled as their constrainees in C#,
            # so we do not transform them.
            continue

        elif isinstance(symbol, intermediate.AbstractClass):
            # Abstract classes are modeled as interfaces in C#, so we do not transform
            # them.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(symbol.name)
            blocks.append(
                Stripped(f"public abstract T Transform({cls_name} that, C context);")
            )

        else:
            assert_never(symbol)

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Perform double-dispatch to transform recursively
/// the instances into something else.
/// </summary>
/// <typeparam name="C">The type of the transformation context</typeparam>
/// <typeparam name="T">The type of the transformation result</typeparam>
public abstract class AbstractTransformerWithContext<C, T>
{I}: ITransformerWithContext<C, T>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public abstract class AbstractTransformerWithContext")

    return Stripped(writer.getvalue())


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, namespace: csharp_common.NamespaceIdentifier
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code of the visitors based on the intermediate representation

    The ``namespace`` defines the AAS C# namespace.
    """
    blocks = [csharp_common.WARNING]  # type: List[Rstripped]

    writer = io.StringIO()
    writer.write(f"namespace {namespace}\n{{\n")
    writer.write(f"{I}public static class Visitation\n" f"{I}{{\n")

    visitation_blocks = [
        _generate_ivisitor(symbol_table=symbol_table),
        _generate_visitor_through(symbol_table=symbol_table),
        _generate_abstract_visitor(symbol_table=symbol_table),
        _generate_ivisitor_with_context(symbol_table=symbol_table),
        _generate_abstract_visitor_with_context(symbol_table=symbol_table),
        _generate_itransformer(symbol_table=symbol_table),
        _generate_abstract_transformer(symbol_table=symbol_table),
        _generate_itransformer_with_context(symbol_table=symbol_table),
        _generate_abstract_transformer_with_context(symbol_table=symbol_table),
    ]

    for i, visitation_block in enumerate(visitation_blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(visitation_block, II))

    writer.write(f"\n{I}}}  // public static class Visitation")
    writer.write(f"\n}}  // namespace {namespace}")

    blocks.append(Stripped(writer.getvalue()))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



# pylint: disable=missing-docstring

import textwrap
import unittest

import tests.common
import tests.infer_for_schema.common
from aas_core_codegen import infer_for_schema


class Test_expected(unittest.TestCase):
    def test_no_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_constrained_primitive(str):
                pass


            class Something:
                some_property: Some_constrained_primitive

                def __init__(self, some_property: Some_constrained_primitive) -> None:
                    self.some_property = some_property


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        (
            _,
            something_cls,
            constraints_by_class,
        ) = tests.infer_for_schema.common.parse_to_symbol_table_and_something_cls_and_constraints_by_class(
            source=source
        )

        constraints_by_props = constraints_by_class[something_cls]
        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                ConstraintsByProperty(
                  len_constraints_by_property={},
                  patterns_by_property={})"""
            ),
            text,
        )

    def test_single_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                prefix = "something"
                return match(f"{prefix}-[a-zA-Z]+", text) is not None


            @invariant(lambda self: is_something(self))
            class Some_constrained_primitive(str):
                pass

            class Something:
                some_property: Some_constrained_primitive

                def __init__(self, some_property: Some_constrained_primitive) -> None:
                    self.some_property = some_property


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        (
            _,
            something_cls,
            constraints_by_class,
        ) = tests.infer_for_schema.common.parse_to_symbol_table_and_something_cls_and_constraints_by_class(
            source=source
        )

        constraints_by_props = constraints_by_class[something_cls]
        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                ConstraintsByProperty(
                  len_constraints_by_property={},
                  patterns_by_property={
                    'some_property': [
                      PatternConstraint(
                        pattern='something-[a-zA-Z]+')]})"""
            ),
            text,
        )

    def test_two_patterns(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None


            @invariant(lambda self: is_acme(self))
            @invariant(lambda self: is_something(self))
            class Some_constrained_primitive(str):
                pass

            class Something:
                some_property: Some_constrained_primitive

                def __init__(self, some_property: Some_constrained_primitive) -> None:
                    self.some_property = some_property


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        (
            _,
            something_cls,
            constraints_by_class,
        ) = tests.infer_for_schema.common.parse_to_symbol_table_and_something_cls_and_constraints_by_class(
            source=source
        )

        constraints_by_props = constraints_by_class[something_cls]
        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                ConstraintsByProperty(
                  len_constraints_by_property={},
                  patterns_by_property={
                    'some_property': [
                      PatternConstraint(
                        pattern='something-[a-zA-Z]+'),
                      PatternConstraint(
                        pattern='.*acme.*')]})"""
            ),
            text,
        )

    def test_inheritance_between_constrained_primitives_by_default(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None

            @invariant(lambda self: is_something(self))
            class Parent_constrained_primitive(str):
                pass

            @invariant(lambda self: is_acme(self))
            class Some_constrained_primitive(Parent_constrained_primitive):
                pass

            class Something:
                some_property: Some_constrained_primitive

                def __init__(self, some_property: Some_constrained_primitive) -> None:
                    self.some_property = some_property


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        # NOTE (mristin, 2022-05-15):
        # In contrast to classes, we do inherit the constraints among the constrained
        # primitives as we in-line them later in the schema classes.

        (
            _,
            something_cls,
            constraints_by_class,
        ) = tests.infer_for_schema.common.parse_to_symbol_table_and_something_cls_and_constraints_by_class(
            source=source
        )

        constraints_by_props = constraints_by_class[something_cls]
        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                ConstraintsByProperty(
                  len_constraints_by_property={},
                  patterns_by_property={
                    'some_property': [
                      PatternConstraint(
                        pattern='something-[a-zA-Z]+'),
                      PatternConstraint(
                        pattern='.*acme.*')]})"""
            ),
            text,
        )


if __name__ == "__main__":
    unittest.main()



"""Merge constrained primitives as property constraints."""
import collections
from typing import Tuple, Optional, List, Mapping, MutableMapping, Sequence

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error
from aas_core_codegen.infer_for_schema import (
    _len as infer_for_schema_len,
    _pattern as infer_for_schema_pattern,
)
from aas_core_codegen.infer_for_schema._types import (
    ConstraintsByProperty,
    LenConstraint,
    PatternConstraint,
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _infer_len_constraints_by_constrained_primitive(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[
    Optional[MutableMapping[intermediate.ConstrainedPrimitive, LenConstraint]],
    Optional[List[Error]],
]:
    """Infer the constraints on ``len(.)`` of the constrained primitives."""

    # NOTE (mristin, 2022-02-11):
    # We do this inference in two passes. In the first pass, we only infer
    # the constraints defined for the constrained primitive and ignore the ancestors.
    # In the second pass, we stack the constraints of the ancestors as well.

    errors = []  # type: List[Error]

    first_pass: MutableMapping[
        intermediate.ConstrainedPrimitive, LenConstraint
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.ConstrainedPrimitive):
            (
                len_constraint,
                len_constraint_errors,
            ) = infer_for_schema_len.infer_len_constraint_of_self(
                constrained_primitive=symbol
            )

            if len_constraint_errors is not None:
                errors.extend(len_constraint_errors)
            else:
                assert len_constraint is not None

                first_pass[symbol] = len_constraint

    if len(errors) > 0:
        return None, errors

    second_pass: MutableMapping[
        intermediate.ConstrainedPrimitive, LenConstraint
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols_topologically_sorted:
        if isinstance(symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-02-11):
            # We make the copy in order to avoid bugs when we start processing
            # the inheritances.
            len_constraint = first_pass[symbol].copy()

            for inheritance in symbol.inheritances:
                inherited_len_constraint = second_pass.get(inheritance, None)
                assert (
                    inherited_len_constraint is not None
                ), "Expected topological order"

                if inherited_len_constraint.min_value is not None:
                    len_constraint.min_value = (
                        max(
                            len_constraint.min_value, inherited_len_constraint.min_value
                        )
                        if len_constraint.min_value is not None
                        else inherited_len_constraint.min_value
                    )

                if inherited_len_constraint.max_value is not None:
                    len_constraint.max_value = (
                        min(
                            len_constraint.max_value, inherited_len_constraint.max_value
                        )
                        if len_constraint.max_value is not None
                        else inherited_len_constraint.max_value
                    )

            second_pass[symbol] = len_constraint

    assert len(errors) == 0
    return second_pass, None


def _infer_pattern_constraints_by_constrained_primitive(
    symbol_table: intermediate.SymbolTable,
    pattern_verifications_by_name: infer_for_schema_pattern.PatternVerificationsByName,
) -> MutableMapping[intermediate.ConstrainedPrimitive, List[PatternConstraint]]:
    """Infer the pattern constraints of the constrained strings."""

    # NOTE (mristin, 2022-02-11):
    # We do this inference in two passes. In the first pass, we only infer
    # the constraints defined for the constrained primitive and ignore the ancestors.
    # In the second pass, we stack the constraints of the ancestors as well.

    first_pass: MutableMapping[
        intermediate.ConstrainedPrimitive,
        List[PatternConstraint],
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols:
        if (
            isinstance(symbol, intermediate.ConstrainedPrimitive)
            and symbol.constrainee is intermediate.PrimitiveType.STR
        ):
            pattern_constraints = infer_for_schema_pattern.infer_patterns_on_self(
                constrained_primitive=symbol,
                pattern_verifications_by_name=pattern_verifications_by_name,
            )

            first_pass[symbol] = pattern_constraints

    second_pass: MutableMapping[
        intermediate.ConstrainedPrimitive,
        List[PatternConstraint],
    ] = collections.OrderedDict()

    for symbol in first_pass:
        # NOTE (mristin, 2022-02-11):
        # We make the copy in order to avoid bugs when we start processing
        # the inheritances.
        pattern_constraints = first_pass[symbol][:]

        for inheritance in symbol.inheritances:
            assert inheritance in first_pass, (
                f"We are processing the constrained primitive {symbol.name!r}. "
                f"However, its parent, {inheritance.name!r}, has not been processed in "
                f"the first pass. Something probably went wrong in the first pass."
            )

            inherited_pattern_constraints = second_pass.get(inheritance, None)
            assert inherited_pattern_constraints is not None, (
                f"Expected topological order. However, the symbol {symbol.name!r} "
                f"is being processed before one of its parents, {inheritance.name!r}."
            )

            pattern_constraints = inherited_pattern_constraints + pattern_constraints

        second_pass[symbol] = pattern_constraints

    return second_pass


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_constraints_by_class(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[
    Optional[MutableMapping[intermediate.ClassUnion, ConstraintsByProperty]],
    Optional[List[Error]],
]:
    """Infer the constraints from the invariants and constrained primitives."""
    errors = []  # type: List[Error]

    pattern_verifications_by_name = (
        infer_for_schema_pattern.map_pattern_verifications_by_name(
            verifications=symbol_table.verification_functions
        )
    )

    (
        len_constraints_by_constrained_primitive,
        some_errors,
    ) = _infer_len_constraints_by_constrained_primitive(symbol_table=symbol_table)
    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert len_constraints_by_constrained_primitive is not None

    patterns_by_constrained_primitive = (
        _infer_pattern_constraints_by_constrained_primitive(
            symbol_table=symbol_table,
            pattern_verifications_by_name=pattern_verifications_by_name,
        )
    )

    result: MutableMapping[
        intermediate.ClassUnion, ConstraintsByProperty
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols:
        if not isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            continue

        # region Infer constraints on ``len(.)``

        len_constraints_by_property: MutableMapping[
            intermediate.Property, LenConstraint
        ] = collections.OrderedDict()

        (
            len_constraints_from_invariants,
            len_constraints_errors,
        ) = infer_for_schema_len.len_constraints_from_invariants(cls=symbol)

        if len_constraints_errors is not None:
            errors.extend(len_constraints_errors)
            continue

        assert len_constraints_from_invariants is not None

        patterns_by_property: MutableMapping[
            intermediate.Property, List[PatternConstraint]
        ] = collections.OrderedDict()

        patterns_from_invariants_by_property = (
            infer_for_schema_pattern.patterns_from_invariants(
                cls=symbol, pattern_verifications_by_name=pattern_verifications_by_name
            )
        )

        # region Merge the length constraints

        for prop in symbol.properties:
            # NOTE (mristin, 2022-03-03):
            # We need to go beneath ``Optional`` as the constraints are applied even
            # if a property is optional. In cases where cardinality is affected by
            # ``Optional``, the client code needs to cover them separately.
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            len_constraint_from_type: Optional[LenConstraint] = None

            len_constraint_from_invariants = len_constraints_from_invariants.get(
                prop, None
            )

            if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.symbol, intermediate.ConstrainedPrimitive
            ):
                len_constraint_from_type = len_constraints_by_constrained_primitive.get(
                    type_anno.symbol, None
                )

            # Merge the constraint from the type and from the invariants

            if (
                len_constraint_from_type is None
                and len_constraint_from_invariants is None
            ):
                pass

            elif (
                len_constraint_from_type is not None
                and len_constraint_from_invariants is None
            ):
                if (
                    len_constraint_from_type.min_value is not None
                    or len_constraint_from_type.max_value is not None
                ):
                    len_constraints_by_property[prop] = len_constraint_from_type

            elif (
                len_constraint_from_type is None
                and len_constraint_from_invariants is not None
            ):
                if (
                    len_constraint_from_invariants.min_value is not None
                    or len_constraint_from_invariants.max_value is not None
                ):
                    len_constraints_by_property[prop] = len_constraint_from_invariants

            elif (
                len_constraint_from_type is not None
                and len_constraint_from_invariants is not None
            ):
                # NOTE (mristin, 2022-03-02):
                # We have to make the bounds *stricter* since both
                # the type constraints and the invariant(s) need to be satisfied.

                min_value = infer_for_schema_len.max_with_none(
                    len_constraint_from_type.min_value,
                    len_constraint_from_invariants.min_value,
                )

                max_value = infer_for_schema_len.min_with_none(
                    len_constraint_from_type.max_value,
                    len_constraint_from_invariants.max_value,
                )

                if (
                    min_value is not None
                    and max_value is not None
                    and min_value > max_value
                ):
                    errors.append(
                        Error(
                            symbol.parsed.node,
                            f"The inferred minimum and maximum value on len(.) "
                            f"is contradictory: "
                            f"minimum = {min_value}, maximum = {max_value}; "
                            f"please check the invariants and "
                            f"any involved constrained primitives",
                        )
                    )
                    continue

                if min_value is not None or max_value is not None:
                    len_constraints_by_property[prop] = LenConstraint(
                        min_value=min_value, max_value=max_value
                    )

            else:
                raise AssertionError(
                    f"Unhandled case: "
                    f"{len_constraint_from_type=}, {len_constraint_from_invariants}"
                )

        # endregion

        # region Infer constraints on string patterns

        for prop in symbol.properties:
            # NOTE (mristin, 2022-03-03):
            # We need to go beneath ``Optional`` as the constraints are applied even
            # if a property is optional. In cases where cardinality is affected by
            # ``Optional``, the client code needs to cover them separately.
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            patterns_from_type: List[PatternConstraint] = []
            patterns_from_invariants = patterns_from_invariants_by_property.get(
                prop, []
            )

            if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.symbol, intermediate.ConstrainedPrimitive
            ):
                patterns_from_type = patterns_by_constrained_primitive.get(
                    type_anno.symbol, []
                )

            merged = patterns_from_type + patterns_from_invariants

            if len(merged) > 0:
                patterns_by_property[prop] = merged

        # endregion

        result[symbol] = ConstraintsByProperty(
            len_constraints_by_property=len_constraints_by_property,
            patterns_by_property=patterns_by_property,
        )

    if len(errors) > 0:
        return None, errors

    return result, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def merge_constraints_with_ancestors(
    symbol_table: intermediate.SymbolTable,
    constraints_by_class: Mapping[intermediate.ClassUnion, ConstraintsByProperty],
) -> Tuple[
    Optional[MutableMapping[intermediate.ClassUnion, ConstraintsByProperty]],
    Optional[Error],
]:
    """
    Merge the constraints over all the classes with their ancestors.

    Usually, when you generate a schema, you do *not* want to inherit the constraints
    over the properties. Most schema engines will do that for you and you want to be
    as explicit as possible in the schema for readability (whereas merged constraints
    might not be as readable, since you do not explicitly see their origin).

    However, for some applications we indeed want to stack the constraints and merge
    them. For example, this is the case when we (semi-)automatically generate test
    data. In those cases, you should use this function.

    The length constraints are merged by picking the smaller interval that fits.
    Patterns are simply stacked together.
    """
    new_constraints_by_class: MutableMapping[
        intermediate.ClassUnion, ConstraintsByProperty
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols_topologically_sorted:
        if not isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            continue

        this_constraints_by_props = constraints_by_class[symbol]

        new_len_constraints_by_property: MutableMapping[
            intermediate.Property, LenConstraint
        ] = collections.OrderedDict()

        new_patterns_by_property: MutableMapping[
            intermediate.Property, Sequence[PatternConstraint]
        ] = collections.OrderedDict()

        for prop in symbol.properties:
            # region Merge len constraints

            len_constraints = []

            this_len_constraint = (
                this_constraints_by_props.len_constraints_by_property.get(prop, None)
            )
            if this_len_constraint is not None:
                len_constraints.append(this_len_constraint)

            for parent in symbol.inheritances:
                # NOTE (mristin, 2022-05-15):
                # Assume here that all the ancestors already inherited their constraints due to
                # the topological order in the iteration.
                that_constraints_by_props = new_constraints_by_class[parent]

                that_len_constraint = (
                    that_constraints_by_props.len_constraints_by_property.get(
                        prop, None
                    )
                )
                if that_len_constraint is not None:
                    len_constraints.append(that_len_constraint)

            min_value = None
            max_value = None

            for len_constraint in len_constraints:
                if min_value is None:
                    min_value = len_constraint.min_value
                else:
                    if len_constraint.min_value is not None:
                        min_value = max(len_constraint.min_value, min_value)

                if max_value is None:
                    max_value = len_constraint.max_value
                else:
                    if len_constraint.max_value is not None:
                        max_value = min(len_constraint.max_value, max_value)

            if (
                min_value is not None
                and max_value is not None
                and min_value > max_value
            ):
                return None, Error(
                    symbol.parsed.node,
                    f"We could not stack the length constraints "
                    f"on the property {prop.name} as they are contradicting: "
                    f"min_value == {min_value} and max_value == {max_value}. "
                    f"Please check the invariants and the invariants of all "
                    f"the ancestors.",
                )

            if min_value is not None or max_value is not None:
                new_len_constraints_by_property[prop] = LenConstraint(
                    min_value=min_value, max_value=max_value
                )

            # endregion

            # region Merge patterns

            # NOTE (mristin, 2022-05-15):
            # The following logic has quadratic time complexity, but it seems that
            # the runtime is currently no problem in practice.

            patterns = []  # type: List[PatternConstraint]

            this_patterns = this_constraints_by_props.patterns_by_property.get(
                prop, None
            )
            if this_patterns is not None:
                patterns.extend(this_patterns)

            for parent in symbol.inheritances:
                # NOTE (mristin, 2022-05-15):
                # Assume here that all the ancestors already inherited their constraints due to
                # the topological order in the iteration.
                that_constraints_by_props = new_constraints_by_class[parent]

                that_patterns = that_constraints_by_props.patterns_by_property.get(
                    prop, None
                )

                if that_patterns is not None:
                    patterns.extend(that_patterns)

            if len(patterns) > 0:
                new_patterns_by_property[prop] = patterns

            # endregion

        new_constraints_by_class[symbol] = ConstraintsByProperty(
            len_constraints_by_property=new_len_constraints_by_property,
            patterns_by_property=new_patterns_by_property,
        )

    return new_constraints_by_class, None



# pylint: disable=missing-docstring

import textwrap
import unittest.mock

# noinspection PyProtectedMember
import aas_core_codegen.csharp.description as csharp_description
import tests.common
from aas_core_codegen.common import Stripped, Identifier


class Test_to_render_meta_model_description(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """Generate the C# description comment based on ``source``."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        assert (
            symbol_table.meta_model.description is not None
        ), "Expected a meta-model description, but found none."

        code, errors = csharp_description.generate_meta_model_comment(
            symbol_table.meta_model.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            """"""  # Intentionally left empty

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)
        assert error is not None
        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_only_summary(self) -> None:
        comment_code = Test_to_render_meta_model_description.render(
            textwrap.dedent(
                '''\
                """Do & drink something."""

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>"""
            ),
            comment_code,
        )


class Test_to_render_symbol_description(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """
        Generate the C# description comment based on ``source``.

        The ``source`` is expected to contain only a single class, a class
        ``Some_class``.
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_class = symbol_table.must_find(Identifier("Some_class"))
        assert some_class.description is not None

        code, errors = csharp_description.generate_symbol_comment(
            some_class.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """"""  # Intentionally left empty

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)
        assert error is not None
        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_no_summary(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                * Some
                * Bullet
                * List
                """

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Expected the first document element to be a summary and thus a paragraph, "
            "but got: "
            '<bullet_list bullet="*">'
            "<list_item><paragraph>Some</paragraph></list_item>"
            "<list_item><paragraph>Bullet</paragraph></list_item>"
            "<list_item><paragraph>List</paragraph></list_item>"
            "</bullet_list>",
            tests.common.most_underlying_messages(error),
        )

    def test_unexpected_directive(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                Do something.

                :someDirective:
                    I am unexpected.
                """

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Expected only directives such as ``constraint some-identifier`` "
            "in this context, but got a directive with 1 part(s): someDirective",
            tests.common.most_underlying_messages(error),
        )

    def test_only_summary(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """Do & drink something."""

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_class_reference(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """Do & drink :class:`.Some_class`."""

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="SomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_interface_reference(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                @abstract
                class Some_class:
                    """Do & drink :class:`.Some_class`."""

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="ISomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_enumeration_reference(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                class Some_class(Enum):
                    """Do & drink :class:`.Some_class`."""

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="SomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_and_remarks(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """
                    Do & drink something.

                    First & remark.

                    Second & remark.
                    """

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>
                /// <remarks>
                /// First &amp; remark.
                ///
                /// Second &amp; remark.
                /// </remarks>"""
            ),
            comment_code,
        )

    def test_summary_remarks_and_constraints(self) -> None:
        comment_code = Test_to_render_symbol_description.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """
                    Do & drink something.

                    First & remark.

                    :constraint AAS-001:
                        You have to do something.
                    """

                __book_url__ = "dummy"
                __book_version__ = "dummy"
                '''
            )
        )
        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>
                /// <remarks>
                /// First &amp; remark.
                ///
                /// Constraints:
                /// <ul>
                ///     <li>
                ///     Constraint AAS-001:
                ///     You have to do something.
                ///     </li>
                /// </ul>
                /// </remarks>"""
            ),
            comment_code,
        )


class Test_to_render_signature_description(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """
        Generate the C# description comment based on ``source``.

        The ``source`` is expected to contain only a single verification function,
        ``verify_something``.
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        verification = symbol_table.verification_functions_by_name[
            Identifier("verify_something")
        ]

        assert verification.description is not None

        code, errors = csharp_description.generate_signature_comment(
            verification.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(text: str) -> bool:
                """"""  # Intentionally left empty
                return match(r'.*', text) is not None

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_only_summary(self) -> None:
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(text: str) -> bool:
                """Verify something."""
                return match(r'.*', text) is not None

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        code = Test_to_render_signature_description.render(source=source)

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Verify something.
                /// </summary>"""
            ),
            code,
        )

    def test_params_and_returns(self) -> None:
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(text: str) -> bool:
                """
                Verify something.

                :param text: to be checked
                :returns: True if :paramref:`text` is a valid something.
                """
                return match(r'.*', text) is not None

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        code = Test_to_render_signature_description.render(source=source)

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Verify something.
                /// </summary>
                /// <param name="text">
                /// to be checked
                /// </param>
                /// <returns>
                /// True if <paramref name="text" /> is a valid something.
                /// </returns>"""
            ),
            code,
        )


if __name__ == "__main__":
    unittest.main()



"""Provide functions shared among the tests."""
from typing import Tuple, MutableMapping

import tests.common
from aas_core_codegen import intermediate, infer_for_schema
from aas_core_codegen.common import Identifier


def parse_to_symbol_table_and_something_cls(
    source: str,
) -> Tuple[intermediate.SymbolTable, intermediate.ClassUnion]:
    """
    Parse the ``source``.

    Return the symbol table and the symbol corresponding to the class ``Something``
    in the ``source``.
    """
    symbol_table, error = tests.common.translate_source_to_intermediate(source=source)
    assert error is None, tests.common.most_underlying_messages(error)
    assert symbol_table is not None
    something_cls = symbol_table.must_find(Identifier("Something"))
    assert isinstance(
        something_cls, (intermediate.AbstractClass, intermediate.ConcreteClass)
    )

    return symbol_table, something_cls


def parse_to_symbol_table_and_something_cls_and_constraints_by_class(
    source: str,
) -> Tuple[
    intermediate.SymbolTable,
    intermediate.ClassUnion,
    MutableMapping[intermediate.ClassUnion, infer_for_schema.ConstraintsByProperty],
]:
    """
    Parse the ``source``.

    Return the symbol table and the symbol corresponding to the class ``Something``
    in the ``source`` as well as the the inferred constraints mapped by classes.
    """
    symbol_table, something_cls = parse_to_symbol_table_and_something_cls(source=source)

    constraints_by_class, error = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )
    assert error is None, tests.common.most_underlying_messages(error)
    assert constraints_by_class is not None

    return symbol_table, something_cls, constraints_by_class
